# -*- coding: utf-8 -*-
#
# GNU General Public License (GPL)
#

from AccessControl.SecurityManagement import getSecurityManager
from collections import OrderedDict
from collective.contact.plonegroup.config import ORGANIZATIONS_REGISTRY
from collective.contact.plonegroup.utils import get_own_organization
from collective.contact.plonegroup.utils import get_plone_groups
from collective.documentviewer.config import CONVERTABLE_TYPES
from collective.documentviewer.settings import GlobalSettings
from collective.iconifiedcategory.utils import calculate_category_id
from collective.iconifiedcategory.utils import get_config_root
from collective.iconifiedcategory.utils import get_group
from copy import deepcopy
from datetime import datetime
from imio.actionspanel.utils import unrestrictedRemoveGivenObject
from imio.helpers.cache import cleanRamCache
from imio.helpers.cache import cleanRamCacheFor
from imio.helpers.content import get_vocab_values
from imio.helpers.content import object_values
from imio.helpers.content import richtextval
from imio.helpers.testing import testing_logger
from imio.helpers.workflow import get_transitions
from plone import api
from plone import namedfile
from plone.app.testing import login
from plone.app.testing import logout
from plone.app.testing.bbb import _createMemberarea
from plone.app.testing.helpers import setRoles
from plone.dexterity.utils import createContentInContainer
from plone.dexterity.utils import iterSchemata
from Products.Archetypes.event import ObjectEditedEvent
from Products.CMFPlone.utils import base_hasattr
from Products.Five.browser import BrowserView
from Products.PloneMeeting.browser.meeting import get_default_attendees
from Products.PloneMeeting.browser.meeting import get_default_signatories
from Products.PloneMeeting.browser.meeting import get_default_voters
from Products.PloneMeeting.config import DEFAULT_USER_PASSWORD
from Products.PloneMeeting.config import ITEM_DEFAULT_TEMPLATE_ID
from Products.PloneMeeting.config import ITEM_SCAN_ID_NAME
from Products.PloneMeeting.config import TOOL_FOLDER_ANNEX_TYPES
from Products.PloneMeeting.testing import PM_TESTING_PROFILE_FUNCTIONAL
from Products.PloneMeeting.tests.helpers import PloneMeetingTestingHelpers
from Products.PloneMeeting.utils import cleanMemoize
from z3c.form.testing import TestRequest as z3c_form_TestRequest
from z3c.relationfield.relation import RelationValue
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.event import notify
from zope.interface import Invalid
from zope.intid.interfaces import IIntIds
from zope.lifecycleevent import ObjectModifiedEvent
from zope.traversing.interfaces import BeforeTraverseEvent
from zope.viewlet.interfaces import IViewletManager

import os.path
import Products.PloneMeeting
import transaction
import unittest


# Force application logging level to DEBUG so we can use logger in tests
pm_logger = testing_logger('PloneMeeting: testing')

IMG_BASE64_DATA = "data:image/gif;base64,R0lGODlhCgAKAPcAAP////79/f36+/3z8/zy8/rq7Prm6Pnq7P" \
    "je4vTx8vPg5O6gqe2gqOq4v+igqt9tetxSYNs7Tdo5TNc5TNUbMdUbMNQRKNIKIdIJINIGHdEGHtEDGtACFdAA" \
    "FtAAFNAAEs8AFAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA" \
    "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA" \
    "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA" \
    "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA" \
    "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA" \
    "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA" \
    "AAAAAAAAAAAAAAAAAAAAAAAAAAAAACIAIP3sAAoACBL2OAAAAPjHWRQAABUKiAAAABL2FAAAAhL+dPkBhPm4MP" \
    "///xL2YPZNegAAAAAQAAAABgAAAAAAABL2wAAAAAAARRL25PEPfxQAAPEQAAAAAAAAA/3r+AAAAAAAGAAAABL2" \
    "uAAAQgAAABL2pAAAAAAAAAAAAAAADAAAAgABAfDTAAAAmAAAAAAAAAAAABL27Mku0AAQAAAAAAAAAEc1MUaDsA" \
    "AAGBL3FPELyQAAAAABEgAAAwAAAQAAAxL22AAAmBL+dPO4dPPKQP///0c70EUoOQAAmMku0AAQABL3TAAAAEaD" \
    "sEc1MUaDsAAABkT98AABEhL3wAAALAAAQAAAQBL3kQAACSH5BAEAAAAALAAAAAAKAAoAQAhGAAEIHEhw4IIIFC" \
    "AsKCBwQAQQFyKCeKDAIEKFDAE4hBiRA0WCAwwUBLCAA4cMICg0YIjAQoaIFzJMSCCw5EkOKjM2FDkwIAA7"


class Response:
    def reset(self):
        pass

    def handleException(self, *args):
        pass

    def redirect(self, *args):
        pass


class TestRequest(z3c_form_TestRequest):

    response = Response()
    RESPONSE = Response()

    def __setitem__(self, attr, value):
        """ """
        self.__setattr__(attr, value)

    def __getitem__(self, attr):
        return self.__getattr__(attr)


class TestFile:
    '''Stub class that simulates a file upload from a HTTP POST.'''
    def __init__(self, testFile, filename):
        self.file = testFile
        self.filename = filename
        self.headers = None


class DefaultData(object):
    """Class used to be passed to a default method."""
    def __init__(self, context):
        self.context = context


class PloneMeetingTestCase(unittest.TestCase, PloneMeetingTestingHelpers):
    '''Base class for defining PloneMeeting test cases.'''

    # Some default content
    descriptionText = '<p>Some description</p>'
    motivationText = '<p>Some motivation.</p>'
    decisionText = '<p>Some decision.</p>'
    subproductIgnoredTestFiles = ['testPerformances.py',
                                  'test_robot.py']

    layer = PM_TESTING_PROFILE_FUNCTIONAL

    cfg1_id = 'plonemeeting-assembly'
    cfg2_id = 'plonegov-assembly'

    external_image1 = \
        "https://fastly.picsum.photos/id/22/400/400.jpg?hmac=Id8VAtx7v59BrMxVGFbHMrf-93mskILQzmMJ__Tzww8"
    external_image2 = \
        "https://fastly.picsum.photos/id/1025/400/300.jpg?hmac=5qUnaqytITcD06pLxsGw7l_twswo9b9p9c8zz_tdpMc"
    external_image3 = \
        "https://fastly.picsum.photos/id/1035/600/400.jpg?hmac=mnooh0fwG-2MIGW-xTUcYO6wyyx9LNdZK4RM6R2SA7A"
    external_image4 = \
        "https://fastly.picsum.photos/id/1062/600/500.jpg?hmac=ZoUBWDuRcsyqDbBPOj5jEU1kHgJ5iGO1edk1-QYode8"

    def setUp(self):
        # enable full diff in failing tests
        self.maxDiff = None
        # Define some useful attributes
        self.app = self.layer['app']
        self.portal = self.layer['portal']
        self.request = self.layer['request']
        self.changeUser('admin')
        # setup manually the correct browserlayer, see:
        # https://dev.plone.org/ticket/11673
        notify(BeforeTraverseEvent(self.portal, self.request))
        self.tool = self.portal.portal_plonemeeting
        self.catalog = self.portal.portal_catalog
        self.wfTool = self.portal.portal_workflow
        self.own_org = get_own_organization()
        # make organizations easily available through their id and store uid
        # for each organization, we will have self.developers, self.developers_uid
        # as well as every plone groups : self.vendors_creators, self.developers_reviewers, ...
        # include organizations outside own_org as well
        self.proposing_groups = object_values(self.own_org, 'PMOrganization')
        self.all_org = object_values(self.own_org.aq_parent, 'PMOrganization') + self.proposing_groups
        self.active_proposing_groups = [org for org in self.proposing_groups if org.active]
        self.inactive_proposing_groups = [org for org in self.proposing_groups if not org.active]
        orgs = object_values(self.own_org, 'PMOrganization') + \
            object_values(self.own_org.aq_parent, 'PMOrganization')
        for org in orgs:
            org_id = org.getId().replace('-', '_')
            setattr(self, org_id, org)
            setattr(self, '{0}_uid'.format(org_id), org.UID())
            for plone_group_id in get_plone_groups(org.UID(), ids_only=True):
                org_uid, suffix = plone_group_id.split('_')
                setattr(self,
                        '{0}_{1}'.format(org_id, suffix),
                        plone_group_id)
        # make held_position easily available as well
        i = 1
        for person in object_values(self.portal.contacts, 'PMPerson'):
            setattr(self,
                    'hp{0}'.format(i),
                    object_values(person, 'PMHeldPosition')[0])
            setattr(self,
                    'hp{0}_uid'.format(i),
                    object_values(person, 'PMHeldPosition')[0].UID())
            i += 1

        self.pmFolder = os.path.dirname(Products.PloneMeeting.__file__)
        # Disable notifications mechanism. This way, the test suite may be
        # executed even on production sites that contain many real users.
        for cfg in self.tool.objectValues('MeetingConfig'):
            cfg.setMailItemEvents([])
            cfg.setMailMeetingEvents([])
        logout()

        # Set the default meeting config
        self.meetingConfig = getattr(self.tool, self.cfg1_id, None)
        self.meetingConfig2 = getattr(self.tool, self.cfg2_id, None)
        # Set the default file and file type for adding annexes
        self.annexFile = u'FILE.txt'
        self.annexFilePDF = u'file_correct.pdf'
        self.annexFileCorruptedPDF = u'file_errorDuringConversion.pdf'
        self.annexFileType = 'financial-analysis'
        self.annexFileTypeDecision = 'decision-annex'
        self.annexFileTypeAdvice = 'advice-annex'
        self.annexFileTypeMeeting = 'meeting-annex'
        # log current test module and method name
        test_num = self._resultForDoCleanups.testsRun
        test_total = self._resultForDoCleanups.count
        pm_logger.info('Executing [{0}/{1}] {2}:{3}'.format(
            test_num, test_total, self.__class__.__name__, self._testMethodName))
        # necessary for MeetingItem.MeetingItemWorkflowConditions._check_required_data
        self.request.set('imio.actionspanel_portal_cachekey', True)

    def tearDown(self):
        self._cleanExistingTmpAnnexFile()

    def createUser(self, username, roles=['Member'], groups=[]):
        '''Creates a user named p_username with some p_roles.'''
        newUser = api.user.create(
            email='test@test.be',
            username=username,
            password=DEFAULT_USER_PASSWORD,
            roles=[],
            properties={})
        setRoles(self.portal, username, roles)
        for group in groups:
            self._addPrincipalToGroup(username, group)
        _createMemberarea(self.portal, username)
        return newUser

    def setMeetingConfig(self, meetingConfigId):
        '''On which meeting config must we work?'''
        self.meetingConfig = getattr(self.tool, meetingConfigId)

    def setCurrentMeeting(self, meeting):
        '''In utils.py, a method is used to get the currently published object
           in the Plone site. Within this test system, it returns None. This
           method allows to simulate that p_meeting is the currently published
           object.'''
        self.portal.REQUEST['PUBLISHED'] = meeting

    def hasPermission(self, permission, obj):
        '''Checks if the currently logged user has the p_permission on p_obj.
           It is not possible to do it for any user, ie:

           user = self.portal.portal_membership.getMemberById(userId)
           return user.has_permission(permission, obj)

           does not work. So we need to change logged user every time.

           Note that p_obj may be a list of object instead of a single object.
           In this case, the method returns True if the currently logged user
           has p_permission on every object of the list.'''

        # make sure we do not have permission check cache problems...
        self.cleanMemoize()

        sm = getSecurityManager()
        res = True
        if type(obj) in (list, tuple):
            for o in obj:
                res = res and sm.checkPermission(permission, o)
        else:
            res = sm.checkPermission(permission, obj)
        return res

    def changeUser(self, loginName, clean_memoize=True):
        '''Logs out currently logged user and logs in p_loginName.'''
        logout()
        if clean_memoize:
            # necessary to invalidate imio.helpers.patches._listAllowedRolesAndUsers
            # or a catalog query reuses cached allowedRolesAndUsers
            cleanRamCache()
            self.cleanMemoize()
        if loginName == 'admin':
            login(self.app, loginName)
        else:
            login(self.portal, loginName)
        self.member = api.user.get_current()
        self.portal.REQUEST['AUTHENTICATED_USER'] = self.member

    def _generateId(self, ploneFolder):
        '''Generate an id for creating an object in p_ploneFolder.'''
        prefix = 'o'
        i = 1
        gotId = False
        while not gotId:
            res = prefix + str(i)
            if not hasattr(ploneFolder, res):
                gotId = True
            else:
                i += 1
        return res

    def getMeetingFolder(self, meetingConfig=None, userId=None):
        '''Get the meeting folder for the current meeting config.'''
        if not meetingConfig:
            meetingConfig = self.meetingConfig
        return self.tool.getPloneMeetingFolder(meetingConfig.id, userId=userId)

    def create(self,
               objectType,
               folder=None,
               autoAddCategory=True,
               is_classifier=False,
               **attrs):
        '''Creates an instance of a meeting (if p_objectType is 'Meeting') or
           meeting item (if p_objectType is 'MeetingItem') and
           returns the created object. p_attrs is a dict of attributes
           that will be given to invokeFactory.'''
        cfg = self.meetingConfig
        shortName = cfg.getShortName()
        # Some special behaviour occurs if the item to create is
        # a recurring item or an item template
        contentType = objectType
        if objectType == 'MeetingItemRecurring':
            contentType = '%s%s' % (objectType, shortName)
            folder = cfg.recurringitems
        elif objectType == 'MeetingItemTemplate':
            contentType = '%s%s' % (objectType, shortName)
            folder = folder or cfg.itemtemplates
        elif objectType == 'MeetingConfig':
            folder = self.tool
        elif objectType == 'organization':
            folder = folder or self.own_org
        elif objectType == 'person':
            folder = self.portal.contacts
        elif objectType == 'held_position':
            if folder is None:
                raise Exception(
                    'The "folder" parameter must be a person when creating a held_position!')
            if "organization" not in attrs:
                attrs['position'] = self._relation(self.own_org)
        elif objectType == 'meetingcategory':
            if is_classifier:
                folder = cfg.classifiers
            else:
                folder = cfg.categories
        elif objectType == 'ConfigurablePODTemplate':
            folder = cfg.podtemplates
        else:
            contentType = '%s%s' % (objectType, shortName)
            folder = self.getMeetingFolder(cfg)
        # Add some computed attributes
        idInAttrs = 'id' in attrs
        if not idInAttrs:
            attrs.update({'id': self._generateId(folder)})
        if objectType == 'Meeting' and attrs.get('date', None) is None:
            attrs.update({'date': datetime.now()})
        if objectType == 'MeetingItem':
            if 'proposingGroup' not in attrs.keys():
                cleanRamCacheFor('Products.PloneMeeting.ToolPloneMeeting._get_org_uids_for_user')
                proposingGroupUids = self.tool.get_orgs_for_user(suffixes=['creators'])
                if len(proposingGroupUids):
                    attrs.update({'proposingGroup': proposingGroupUids[0]})
        obj = getattr(folder, folder.invokeFactory(contentType, **attrs))
        if objectType == 'Meeting':
            self.setCurrentMeeting(obj)
        elif objectType == 'MeetingItem':
            # optionalAdvisers are not set (???) by invokeFactory...
            if 'optionalAdvisers' in attrs:
                obj.setOptionalAdvisers(attrs['optionalAdvisers'])
            # rich text fields are not set (???) by invokeFactory...
            rich_fields = ['motivation', 'decision', 'decisionSuite', 'decisionEnd', 'votesResult']
            for rich_field in rich_fields:
                if rich_field in attrs:
                    field = obj.getField(rich_field)
                    field.set(obj, attrs[rich_field])
            # define a category for the item if necessary
            if autoAddCategory and \
               'category' in cfg.getUsedItemAttributes() and \
               not obj.getCategory():
                aCategory = cfg.getCategories()[0].getId()
                obj.setCategory(aCategory)
        if hasattr(obj.aq_inner, 'processForm'):
            # at_post_create_script is called by processForm
            # but processForm manage the _at_creation_flag
            # keep id if it was given
            if idInAttrs:
                obj._at_rename_after_creation = False
            obj.processForm()
            if idInAttrs:
                obj._at_rename_after_creation = True
        if objectType == 'Meeting':
            # manage attendees if using it
            usedMeetingAttrs = cfg.getUsedMeetingAttributes()
            if 'attendees' in usedMeetingAttrs:
                default_attendees = get_default_attendees(cfg)
                default_attendees = OrderedDict((
                    (attendee, 'attendee') for attendee in default_attendees))
                signatories = []
                if 'signatories' in usedMeetingAttrs:
                    signatories = get_default_signatories(cfg)
                voters = []
                if cfg.getUseVotes():
                    voters = get_default_voters(cfg)
                obj._do_update_contacts(attendees=default_attendees,
                                        signatories=signatories,
                                        voters=voters)
            # manage default values
            add_form = folder.restrictedTraverse('++add++{0}'.format(obj.portal_type))
            add_form.update()
            for field_name, widget in add_form.form_instance.w.items():
                if widget.value and \
                   not getattr(obj, field_name) and \
                   isinstance(widget.value, (str, unicode)):
                    setattr(obj, field_name, widget.field.fromUnicode(widget.value))

        # make sure we do not have permission check cache problems...
        self.cleanMemoize()
        return obj

    def _cleanExistingTmpAnnexFile(self):
        '''While adding in annex (see code around shutil in addAnnex),
           a temporary file is created.  In case we check assertRaise(Unauthorized, addAnnex, ...)
           the temporary file is not removed, so make sure it is...'''
        originalAnnexPath = os.path.join(self.pmFolder, self.annexFile)
        newAnnexPath = originalAnnexPath[:-4] + '_tmp_for_tests.%s' % originalAnnexPath[-3:]
        if os.path.exists(newAnnexPath):
            os.remove(newAnnexPath)

    def _annex_file_content(self, annexFile=None):
        current_path = os.path.dirname(__file__)
        annexFile = annexFile or self.annexFile
        f = open(os.path.join(current_path, annexFile), 'r')
        annex_file = namedfile.NamedBlobFile(f.read(), filename=annexFile)
        return annex_file

    def addAnnex(self,
                 context,
                 annexType=None,
                 annexTitle=None,
                 annexDescription=None,
                 relatedTo=None,
                 to_print=False,
                 confidential=False,
                 to_sign=False,
                 signed=False,
                 publishable=False,
                 annexFile=None,
                 scan_id=None):
        '''Adds an annex to p_item.
           If no p_annexType is provided, self.annexFileType is used.
           If no p_annexTitle is specified, the predefined title of the annex type is used.'''

        if annexType is None:
            if context.getTagName() == 'MeetingItem':
                if not relatedTo:
                    annexType = self.annexFileType
                elif relatedTo == 'item_decision':
                    annexType = self.annexFileTypeDecision
            elif context.portal_type.startswith('meetingadvice'):
                annexType = self.annexFileTypeAdvice
            elif context.getTagName() == 'Meeting':
                annexType = self.annexFileTypeMeeting

        # get complete annexType id that is like
        # 'meeting-config-id-annexes_types_-_item_annexes_-_financial-analysis'
        if relatedTo == 'item_decision':
            context.REQUEST.set('force_use_item_decision_annexes_group', True)
        annexes_config_root = get_config_root(context)
        if relatedTo == 'item_decision':
            context.REQUEST.set('force_use_item_decision_annexes_group', False)
        annexTypeId = calculate_category_id(annexes_config_root.get(annexType))

        annexContentType = 'annex'
        if relatedTo == 'item_decision':
            annexContentType = 'annexDecision'

        # scan_id is removed by default
        self.request.set(ITEM_SCAN_ID_NAME, scan_id)
        annex = createContentInContainer(
            container=context,
            portal_type=annexContentType,
            title=annexTitle or 'Annex',
            description=annexDescription,
            file=self._annex_file_content(annexFile=annexFile),
            content_category=annexTypeId,
            to_print=to_print,
            confidential=confidential,
            to_sign=to_sign,
            signed=signed,
            publishable=publishable,
            scan_id=scan_id)
        self.request.set(ITEM_SCAN_ID_NAME, None)

        class DummyData(object):
            def __init__(self, context, contentType, content_category):
                self.__context__ = context
                self.contentType = contentType
                self.content_category = content_category

        data = DummyData(annex, annex.file.contentType, annex.content_category)
        for schema in iterSchemata(annex):
            try:
                schema.validateInvariants(data, [])
            except Invalid as exc:
                unrestrictedRemoveGivenObject(annex)
                raise(exc)

        # need to commit the transaction so the stored blob is correct
        # if not done, accessing the blob will raise 'BlobError: Uncommitted changes'
        transaction.commit()
        return annex

    def addAnnexType(self,
                     id,
                     to_print=False,
                     confidential=False,
                     enabled=True,
                     relatedTo='item'):
        cfg = self.meetingConfig
        folder = getattr(cfg, TOOL_FOLDER_ANNEX_TYPES)

        portal_type = 'ContentCategory'
        if relatedTo == 'item':
            portal_type = 'ItemAnnexContentCategory'
            categoryGroupId = 'item_annexes'
        elif relatedTo == 'item_decision':
            portal_type = 'ItemAnnexContentCategory'
            categoryGroupId = 'item_decision_annexes'
        elif relatedTo == 'advice':
            categoryGroupId = 'advice_annexes'
        elif relatedTo == 'meeting':
            categoryGroupId = 'meeting_annexes'

        annexType = api.content.create(
            type=portal_type,
            id=id,
            title=id,
            icon=self._annex_file_content(),
            container=getattr(folder, categoryGroupId),
            to_print=to_print,
            confidential=confidential,
            enabled=enabled
        )
        return annexType

    def addAdvice(self,
                  item,
                  advice_group=None,
                  advice_type=u"positive",
                  advice_comment=u"My comment",
                  advice_hide_during_redaction=False,
                  advice_portal_type='meetingadvice'):
        if not advice_group:
            advice_group = self.vendors_uid
        # manage MeetingConfig.defaultAdviceHiddenDuringRedaction
        # as it only works while added ttw
        if not advice_hide_during_redaction:
            advice_hide_during_redaction = advice_portal_type in \
                self.meetingConfig.getDefaultAdviceHiddenDuringRedaction()
        advice = createContentInContainer(
            item,
            advice_portal_type,
            **{'advice_group': advice_group,
               'advice_type': advice_type,
               'advice_hide_during_redaction': advice_hide_during_redaction,
               'advice_comment': richtextval(advice_comment)})
        return advice

    def deleteAsManager(self, uid):
        """When we want to remove an item the current user does not have permission to,
           but we are not testing delete permission, do it as a 'Manager'."""
        currentUser = self.member.getId()
        self.changeUser('admin')
        self.portal.restrictedTraverse('@@delete_givenuid')(uid)
        self.changeUser(currentUser)

    def cleanMemoize(self):
        """
          Remove every memoized informations
        """
        # borg localroles are memoized...
        # so while checking local roles twice, there could be problems...
        # remove memoized localroles
        cleanMemoize(self.portal,
                     prefixes=['borg.localrole.workspace.checkLocalRolesAllowed',
                               'tool-getmeetinggroups-',
                               'meeting-config-getcategories-',
                               'meeting-config-gettopics-',
                               'plonegroup-utils-get_organizations-',
                               'PloneMeeting-MeetingConfig-getMeetingsAcceptingItems',
                               'PloneMeeting-tool-get_orgs_for_user',
                               'ftw-labeling-cache-'])

    def _removeOrganizations(self):
        """Delete every organizations found in own_org."""
        # remove every users from linked Plone groups so organizations are removable
        for org in self.own_org.objectValues():
            plone_groups = get_plone_groups(org.UID())
            for plone_group in plone_groups:
                for member_id in plone_group.getMemberIds():
                    self.portal.portal_groups.removePrincipalFromGroup(member_id, plone_group.id)
            # empty groups_in_charge
            org.groups_in_charge = []
        # unselect every organizations from plonegroup
        api.portal.set_registry_record(ORGANIZATIONS_REGISTRY, [])
        ids_to_remove = self.own_org.objectIds()
        self.own_org.manage_delObjects(ids=ids_to_remove)
        self.cleanMemoize()

    def _removeConfigObjectsFor(self, meetingConfig, folders=['recurringitems', 'itemtemplates']):
        """
          Helper method for removing some objects from a given p_meetingConfig.
          Given p_folders are folders of the meetingConfig to clean out.
        """
        currentUser = self.member.getId()
        self.changeUser('admin')
        for folderId in folders:
            # folder could be subfolder/subsubfolder
            subfolder = meetingConfig
            for subfolderId in folderId.split('/'):
                subfolder = getattr(subfolder, subfolderId)
            objectIds_to_remove = []
            for obj in subfolder.objectValues():
                # can not remove the ITEM_DEFAULT_TEMPLATE_ID
                if folderId == 'itemtemplates' and \
                   obj.getId() == ITEM_DEFAULT_TEMPLATE_ID:
                    if obj.query_state() == 'active':
                        # disable it instead removing it
                        api.content.transition(obj, 'deactivate')
                    continue
                objectIds_to_remove.append(obj.getId())
            subfolder.manage_delObjects(ids=objectIds_to_remove)
        self.changeUser(currentUser)

    def _turnUserIntoPrereviewer(self, member):
        """
          Helper method for adding a given p_member to every '_prereviewers' group
          corresponding to every '_reviewers' group he is in.
        """
        reviewers = self.meetingConfig.reviewersFor()
        groups = [group for group in member.getGroups() if group.endswith('_%s' % reviewers.keys()[0])]
        groups = [group.replace(reviewers.keys()[0], reviewers.keys()[-1]) for group in groups]
        for group in groups:
            self._addPrincipalToGroup(member.getId(), group)

    # Workflow-related methods -------------------------------------------------
    def do(self, obj, transition, comment='', clean_memoize=True):
        '''Executes a workflow p_transition on a given p_obj.'''
        self.wfTool.doActionFor(obj, transition, comment=comment)
        if clean_memoize:
            self.cleanMemoize()

    def transitions(self, obj):
        '''Returns the list of transition ids that the current user
           may trigger on p_obj.'''
        return sorted(get_transitions(obj))

    def _get_viewlet_manager(self, context, manager_name):
        """ """
        view = BrowserView(context, self.request)
        viewlet_manager = getMultiAdapter(
            (context, self.request, view),
            IViewletManager,
            manager_name)
        viewlet_manager.update()
        return viewlet_manager

    def _get_viewlet(self, context, manager_name, viewlet_name):
        """ """
        viewlet_manager = self._get_viewlet_manager(context, manager_name)
        viewlet = viewlet_manager.get(viewlet_name)
        viewlet.update()
        return viewlet

    def _addPrincipalToGroup(self, principal_id, group_id):
        """We need to changeUser so getGroups is updated."""
        self.portal.portal_groups.addPrincipalToGroup(principal_id, group_id)
        self.changeUser(self.member.getId())

    def _removePrincipalFromGroups(self, principal_id, group_ids):
        """We need to changeUser so getGroups is updated."""
        for group_id in group_ids:
            self.portal.portal_groups.removePrincipalFromGroup(principal_id, group_id)
        self.changeUser(self.member.getId())

    def _setPowerObserverStates(self,
                                cfg=None,
                                observer_type='powerobservers',
                                field_name='item_states',
                                states=[],
                                access_on=''):
        """Change power observers states for item or meeting."""
        if not cfg:
            cfg = self.meetingConfig
        power_observers = deepcopy(cfg.getPowerObservers())
        for po_infos in power_observers:
            if po_infos['row_id'] == observer_type:
                po_infos[field_name] = states
                if field_name == 'item_states':
                    po_infos['item_access_on'] = access_on
                else:
                    po_infos['meeting_access_on'] = access_on
        cfg.setPowerObservers(power_observers)

    def _activate_wfas(self, wfas, cfg=None, keep_existing=False):
        """Activate given p_wfas, we clean wfas, apply,
           then set given p_wfas and apply again."""
        if isinstance(wfas, basestring):
            wfas = [wfas]
        currentUser = self.member.getId()
        self.changeUser('siteadmin')
        if cfg is None:
            cfg = self.meetingConfig
        if not keep_existing:
            cfg.setWorkflowAdaptations(())
            notify(ObjectEditedEvent(cfg))
        else:
            wfas = tuple(set(tuple(wfas) + cfg.getWorkflowAdaptations()))
        if wfas:
            cfg.setWorkflowAdaptations(wfas)
            notify(ObjectEditedEvent(cfg))
        self.changeUser(currentUser)

    def _deactivate_wfas(self, wfas, cfg=None):
        """Deactivate given p_wfas."""
        if isinstance(wfas, basestring):
            wfas = [wfas]
        currentUser = self.member.getId()
        self.changeUser('siteadmin')
        if cfg is None:
            cfg = self.meetingConfig
        wfas = [wfa for wfa in cfg.getWorkflowAdaptations()
                if wfa not in wfas]
        cfg.setWorkflowAdaptations(wfas)
        notify(ObjectEditedEvent(cfg))
        self.changeUser(currentUser)

    def _activate_config(self,
                         field_name,
                         value,
                         cfg=None,
                         keep_existing=True,
                         remove=False,
                         reload=True):
        """Helper to activate a value in the configuration."""
        cfg = cfg or self.meetingConfig
        field = cfg.getField(field_name)
        values = list(field.get(cfg))
        if remove:
            values.remove(value)
        else:
            if keep_existing:
                values.append(value)
            else:
                values = [value]
        field.getMutator(cfg)(values)
        if reload:
            currentUser = self.member.getId()
            self.changeUser('siteadmin')
            notify(ObjectEditedEvent(cfg))
            self.changeUser(currentUser)

    def _enableAutoConvert(self, enable=True):
        """Enable collective.documentviewer auto_convert."""
        gsettings = GlobalSettings(self.portal)
        gsettings.auto_convert = enable
        # False or every portal_type having a file is converted, like PODTemplate, ...
        gsettings.auto_select_layout = False
        gsettings.auto_layout_file_types = CONVERTABLE_TYPES.keys()
        self.tool.at_post_edit_script()
        return gsettings

    def _enable_column(self, column_name, cfg=None, related_to='MeetingItem', enable=True):
        """ """
        cfg = cfg or self.meetingConfig
        column_names = cfg.getItemColumns() if related_to == 'MeetingItem' else cfg.getMeetingColumns()
        if column_name not in column_names:
            column_names += (column_name, )
            if related_to == 'MeetingItem':
                cfg.setItemColumns(column_names)
            else:
                cfg.setMeetingColumns(column_names)
            cfg.updateCollectionColumns()
            return True

    def _enableField(self, field_names, cfg=None, related_to='MeetingItem', enable=True, reload=False):
        """ """
        if not hasattr(field_names, "__iter__"):
            field_names = [field_names]
        cfg = cfg or self.meetingConfig
        for field_name in field_names:
            if related_to == 'MeetingItem':
                usedItemAttrs = list(cfg.getUsedItemAttributes())
                # make sure we are not playing with a field_name that does not exist
                if field_name not in cfg.getField('usedItemAttributes').Vocabulary(cfg):
                    raise Exception("\"%s\" does not exist in usedItemAttributes" % field_name)
                if enable and field_name not in usedItemAttrs:
                    usedItemAttrs.append(field_name)
                elif not enable and field_name in usedItemAttrs:
                    usedItemAttrs.remove(field_name)
                cfg.setUsedItemAttributes(usedItemAttrs)
            elif related_to == 'Meeting':
                usedMeetingAttrs = list(cfg.getUsedMeetingAttributes())
                # make sure we are not playing with a field_name that does not exist
                if field_name not in cfg.getField('usedMeetingAttributes').Vocabulary(cfg):
                    raise Exception("\"%s\" does not exist in usedMeetingAttributes" % field_name)
                if enable and field_name not in usedMeetingAttrs:
                    usedMeetingAttrs.append(field_name)
                elif not enable and field_name in usedMeetingAttrs:
                    usedMeetingAttrs.remove(field_name)
                cfg.setUsedMeetingAttributes(tuple(usedMeetingAttrs))
        if reload:
            currentUser = self.member.getId()
            self.changeUser('siteadmin')
            notify(ObjectEditedEvent(cfg))
            self.changeUser(currentUser)
        else:
            cleanRamCacheFor('Products.PloneMeeting.MeetingItem.attribute_is_used')
            cleanRamCacheFor('Products.PloneMeeting.content.meeting.attribute_is_used')

    def _enable_annex_config(self,
                             obj,
                             param="confidentiality",
                             related_to=None,
                             enable=True):
        """p_fct possible values are :
           - confidentiality (default);
           - to_be_printed;
           - signed;
           - publishable."""
        if related_to == 'item_decision':
            self.request.set('force_use_item_decision_annexes_group', True)
        annexes_config_root = get_config_root(obj)
        if related_to == 'item_decision':
            self.request.set('force_use_item_decision_annexes_group', False)

        annex_group = get_group(annexes_config_root, obj)
        attr_name = "{0}_activated".format(param)
        setattr(annex_group, attr_name, enable)

    def _enable_action(self, action, related_to="MeetingItem", enable=True):
        """Enable an action for given p_related_to element."""
        cfg = self.meetingConfig
        if related_to == "MeetingItem":
            if enable and action not in cfg.getEnabledItemActions():
                actions = cfg.getEnabledItemActions() + (action, )
                cfg.setEnabledItemActions(actions)
                notify(ObjectEditedEvent(cfg))
            elif not enable and action in cfg.getEnabledItemActions():
                actions = list(cfg.getEnabledItemActions())
                actions.remove(action)
                cfg.setEnabledItemActions(actions)
                notify(ObjectEditedEvent(cfg))

    def _disableObj(self, obj, notify_event=True):
        """ """
        # using field 'enabled'
        if base_hasattr(obj, 'enabled'):
            obj.enabled = False
            obj.reindexObject(idxs=['enabled'])
        else:
            # using workflow
            self.do(obj, 'deactivate')
        if notify_event:
            # manage cache
            notify(ObjectModifiedEvent(obj))
        self.cleanMemoize()

    def _check_wfa_available(self, wfas, related_to='MeetingItem'):
        available = True
        if related_to == 'MeetingItem':
            available_wfas = get_vocab_values(self.meetingConfig, 'WorkflowAdaptations')
        elif related_to == 'MeetingAdvice':
            available_wfas = get_vocab_values(self.tool, 'AdviceWorkflowAdaptations')

        for wfa in wfas:
            if wfa not in available_wfas:
                available = False
                pm_logger.info('Bypassing {0} because WFAdaptation {1} is not available!'.format(
                    self._testMethodName, wfa))
                break
        return available

    def validate_at_fields(self, obj):
        """ """
        errors = {}
        for field_name in obj.Schema().keys():
            field = obj.getField(field_name)
            field.validate(field.getAccessor(obj)(), obj, errors)
        return errors

    def _relation(self, obj):
        """ """
        intids = getUtility(IIntIds)
        return RelationValue(intids.getId(obj))



# -*- coding: utf-8 -*-

from collective.contact.plonegroup.browser.settings import IContactPlonegroupConfig
from collective.contact.plonegroup.config import get_registry_functions
from collective.contact.plonegroup.utils import get_all_suffixes
from collective.contact.plonegroup.utils import get_organizations
from collective.contact.plonegroup.utils import get_plone_group_id
from datetime import date
from DateTime import DateTime
from plone import api
from Products.CMFPlone.utils import safe_unicode
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.indexes import REAL_ORG_UID_PATTERN
from Products.PloneMeeting.utils import get_item_validation_wf_suffixes
from Products.PloneMeeting.utils import getInterface
from Products.validation.interfaces.IValidator import IValidator
from z3c.form import validator
from zope.component import getGlobalSiteManager
from zope.component import provideAdapter
from zope.component.hooks import getSite
from zope.i18n import translate
from zope.interface import implements
from zope.interface import Invalid


__author__ = """Gaetan DELANNAY <gaetan.delannay@geezteem.com>, Gauthier BASTIEN
<g.bastien@imio.be>, Stephan GEULETTE <s.geulette@imio.be>"""
__docformat__ = 'plaintext'


def _validate_certified_signatures(value):
    '''Validate the certified signatures format, check that :
       - signatures are sorted by signature number;
       - if dates (date_from and date_to) are provided, both are provided and it respects correct format;
       - 2 lines are not using same 'number/datefrom/dateto'.'''
    lastSignatureNumber = 0
    row_number = 0
    portal = getSite()
    # we will store a "hash" of every signatures so we may check
    # that 2 signatures does not use same number/datefrom/dateto
    signHashes = []
    for signature in value:
        # bypass 'template_row_marker'
        if 'orderindex_' in signature and signature['orderindex_'] == 'template_row_marker':
            continue
        row_number += 1
        # check that signatures are correctly ordered by signature number
        # for now, key may be signatureNumber (MeetingConfig) or signature_number (organization)
        signatureNumber = int(signature.get('signatureNumber', signature.get('signature_number')))
        if signatureNumber < lastSignatureNumber:
            return translate('error_certified_signatures_order',
                             domain='PloneMeeting',
                             context=portal.REQUEST)
        lastSignatureNumber = signatureNumber
        # if a date_from is defined, a date_to is required and vice versa
        # as this work as AT and DX validator, we make sure we have string date
        # because it is the case for AT but for DX, we have datetime.date objects
        date_from = signature['date_from']
        date_to = signature['date_to']
        date_from = isinstance(date_from, date) and date_from.strftime('%Y/%m/%d') or date_from
        date_to = isinstance(date_to, date) and date_to.strftime('%Y/%m/%d') or date_to
        # stop checks if no date provided
        if not date_from and not date_to:
            pass
        else:
            # if a date is provided, both are required
            if (date_from and not date_to) or \
               (date_to and not date_from):
                return translate('error_certified_signatures_both_dates_required',
                                 mapping={'row_number': row_number},
                                 domain='PloneMeeting',
                                 context=portal.REQUEST)
            try:
                datetime_from = DateTime(date_from)
                datetime_to = DateTime(date_to)
                # respect right string format?
                # datefrom <= dateto?
                if not datetime_from.strftime('%Y/%m/%d') == date_from or \
                   not datetime_to.strftime('%Y/%m/%d') == date_to or \
                   not datetime_from <= datetime_to:
                    raise SyntaxError
            except Exception:
                return translate('error_certified_signatures_invalid_dates',
                                 mapping={'row_number': row_number},
                                 domain='PloneMeeting',
                                 context=portal.REQUEST)
        # now check that 2 signatures having same number does not have same period
        # indeed 2 signatures with same number and period is nonsense, the first will still be used
        signHash = "{0}__{1}__{2}".format(signatureNumber, date_from, date_to)
        if signHash in signHashes:
            return translate('error_certified_signatures_duplicated_entries',
                             mapping={'row_number': row_number},
                             domain='PloneMeeting',
                             context=portal.REQUEST)
        signHashes.append(signHash)


class DXCertifiedSignaturesValidator(validator.SimpleFieldValidator):
    """z3c.form validator class for certified signatures format.
    """

    def validate(self, value):
        """
        """
        error = _validate_certified_signatures(value)
        if error:
            raise Invalid(error)


class ATCertifiedSignaturesValidator:
    ''' '''
    implements(IValidator)

    def __init__(self, name, title='', description=''):
        self.name = name
        self.title = title
        self.description = description

    def __call__(self, value, *args, **kwargs):
        """ """
        return _validate_certified_signatures(value)


# Helper class for validating workflow interfaces ------------------------------
WRONG_INTERFACE = 'You must specify here interface "%s" or a subclass of it.'
NO_ADAPTER_FOUND = 'No adapter was found that provides "%s" for "%s".'


class WorkflowInterfacesValidator:
    '''Checks that declared interfaces exist and that adapters were defined for it.'''

    implements(IValidator)

    def __init__(self, baseInterface, baseWorkflowInterface):
        self.baseInterface = baseInterface
        self.baseWorkflowInterface = baseWorkflowInterface

    def _getPackageName(self, klass):
        '''Returns the full package name if p_klass.'''
        return '%s.%s' % (klass.__module__, klass.__name__)

    def __call__(self, value, *args, **kwargs):
        # Get the interface corresponding to the name specified in p_value.
        theInterface = None
        try:
            theInterface = getInterface(value)
        except Exception, e:
            return str(e)
        # Check that this interface is self.baseWorkflowInterface or
        # a subclass of it.
        if not issubclass(theInterface, self.baseWorkflowInterface):
            return WRONG_INTERFACE % (self._getPackageName(
                                      self.baseWorkflowInterface))
        # Check that it exits an adapter that provides theInterface for
        # self.baseInterface.
        sm = getGlobalSiteManager()
        adapter = sm.adapters.lookup1(self.baseInterface, theInterface)
        if not adapter:
            return NO_ADAPTER_FOUND % (self._getPackageName(theInterface),
                                       self._getPackageName(self.baseInterface))


# Complete validation of collective.contact.plonegroup settings "functions"
class PloneGroupSettingsFunctionsValidator(validator.SimpleFieldValidator):

    def validate(self, value):
        # check that if a suffix is removed, it is not used in MeetingConfig or MeetingItems
        stored_suffixes = get_all_suffixes(only_enabled=True)
        # get removed suffixes...
        saved_suffixes = [func['fct_id'] for func in value]
        saved_enabled_suffixes = [func['fct_id'] for func in value if func['enabled']]
        removed_suffixes = list(set(stored_suffixes) - set(saved_enabled_suffixes))
        really_removed_suffixes = list(set(stored_suffixes) - set(saved_suffixes))
        org_uids = get_organizations(only_selected=False, the_objects=False)
        removed_plonegroups = [
            get_plone_group_id(org_uid, removed_suffix)
            for org_uid in org_uids
            for removed_suffix in removed_suffixes]
        # ... and new defined fct_orgs as it will remove some suffixed groups
        stored_functions = get_registry_functions()
        old_functions = {dic['fct_id']: {'fct_title': dic['fct_title'],
                                         'fct_orgs': dic['fct_orgs'],
                                         'enabled': dic['enabled']}
                         for dic in stored_functions}
        new_functions = {dic['fct_id']: {'fct_title': dic['fct_title'],
                                         'fct_orgs': dic['fct_orgs'],
                                         'enabled': dic['enabled']}
                         for dic in value}
        for new_function, new_function_infos in new_functions.items():
            if new_function_infos['fct_orgs'] and \
               old_functions[new_function]['fct_orgs'] != new_function_infos['fct_orgs']:
                # check that Plone group is empty for not selected fct_orgs
                for org_uid in org_uids:
                    if org_uid in new_function_infos['fct_orgs']:
                        continue
                    removed_plonegroups.append(get_plone_group_id(org_uid, new_function))
            elif new_function_infos['enabled'] is False:
                # check that Plone groups are all empty
                for org_uid in org_uids:
                    removed_plonegroups.append(get_plone_group_id(org_uid, new_function))

        # check that plonegroups and suffixes not used in MeetingConfigs
        removed_plonegroups = set(removed_plonegroups)
        tool = api.portal.get_tool('portal_plonemeeting')
        # advisers
        advisers_removed_plonegroups = [
            REAL_ORG_UID_PATTERN.format(removed_plonegroup_id.split('_')[0])
            for removed_plonegroup_id in removed_plonegroups
            if removed_plonegroup_id.endswith('_advisers')]
        for cfg in tool.objectValues('MeetingConfig'):
            msg = _("can_not_delete_plone_group_meetingconfig",
                    mapping={'cfg_title': safe_unicode(cfg.Title(include_config_group=True))})
            # copyGroups
            if removed_plonegroups.intersection(cfg.getSelectableCopyGroups()):
                raise Invalid(msg)
            # advisers (selectableAdvisers/selectableAdviserUsers)
            if set(advisers_removed_plonegroups).intersection(cfg.getSelectableAdvisers()) or \
               set(advisers_removed_plonegroups).intersection(cfg.getSelectableAdviserUsers()):
                raise Invalid(msg)
            # suffixes, values are like 'suffix_proposing_group_level1reviewers'
            composed_values_attributes = ['itemAnnexConfidentialVisibleFor',
                                          'adviceAnnexConfidentialVisibleFor',
                                          'meetingAnnexConfidentialVisibleFor',
                                          'itemInternalNotesEditableBy']
            for composed_values_attr in composed_values_attributes:
                values = cfg.getField(composed_values_attr).getAccessor(cfg)()
                values = [v for v in values
                          for r in removed_suffixes if r in v]
                if values:
                    raise Invalid(msg)
            # itemWFValidationLevels, may be disabled if validation level also disabled
            # but not removed
            item_enabled_val_suffixes = get_item_validation_wf_suffixes(cfg)
            if set(really_removed_suffixes).intersection(item_enabled_val_suffixes):
                raise Invalid(msg)
            all_item_val_suffixes = get_item_validation_wf_suffixes(cfg, only_enabled=False)
            if set(removed_suffixes).intersection(all_item_val_suffixes):
                raise Invalid(msg)
        # check that plone_group not used in MeetingItems
        # need to be performant or may kill the instance when several items exist
        if removed_plonegroups:
            catalog = api.portal.get_tool('portal_catalog')
            # copy_groups
            brains = catalog.unrestrictedSearchResults(
                meta_type="MeetingItem", getCopyGroups=tuple(removed_plonegroups))
            if not brains:
                brains = catalog.unrestrictedSearchResults(
                    meta_type="MeetingItem", indexAdvisers=tuple(advisers_removed_plonegroups))
            for brain in brains:
                item = brain.getObject()
                if item.isDefinedInTool():
                    msgid = "can_not_delete_plone_group_config_meetingitem"
                else:
                    msgid = "can_not_delete_plone_group_meetingitem"
                msg = _(msgid, mapping={'item_url': item.absolute_url()})
                raise Invalid(msg)


validator.WidgetValidatorDiscriminators(
    PloneGroupSettingsFunctionsValidator, field=IContactPlonegroupConfig['functions'])
provideAdapter(PloneGroupSettingsFunctionsValidator)


# Complete validation of collective.contact.plonegroup settings "organizations"
class PloneGroupSettingsOrganizationsValidator(validator.SimpleFieldValidator):

    def validate(self, value):
        selected_org_uids = get_organizations(only_selected=True, the_objects=False)
        removed_org_uids = set(selected_org_uids).difference(value)
        tool = api.portal.get_tool('portal_plonemeeting')
        for cfg in tool.objectValues('MeetingConfig'):
            # usingGroups
            if removed_org_uids.intersection(cfg.getUsingGroups()):
                msg = _("can_not_unselect_plone_group_meetingconfig",
                        mapping={'cfg_title': safe_unicode(cfg.Title(include_config_group=True))})
                raise Invalid(msg)

        # check that removed orgs are not used as groups_in_charge of an organization
        orgs = get_organizations(only_selected=False, the_objects=True)
        for org in orgs:
            if removed_org_uids.intersection(org.groups_in_charge):
                msgid = "can_not_unselect_plone_group_org"
                msg = _(msgid, mapping={'org_url': org.absolute_url()})
                raise Invalid(msg)


validator.WidgetValidatorDiscriminators(
    PloneGroupSettingsOrganizationsValidator, field=IContactPlonegroupConfig['organizations'])
provideAdapter(PloneGroupSettingsOrganizationsValidator)



# -*- coding: utf-8 -*-

from AccessControl import ClassSecurityInfo
from AccessControl import Unauthorized
from collective.contact.plonegroup.utils import get_organization
from dexterity.localrolesfield.field import LocalRoleField
from imio.helpers.content import get_vocab
from imio.helpers.workflow import get_final_states
from imio.helpers.workflow import get_leading_transitions
from imio.history.interfaces import IImioHistory
from imio.history.utils import getLastAction
from imio.history.utils import getLastWFAction
from imio.prettylink.interfaces import IPrettyLink
from plone import api
from plone.app.textfield import RichText
from plone.dexterity.content import Container
from plone.dexterity.schema import DexteritySchemaPolicy
from plone.directives import form
from Products.CMFPlone.utils import base_hasattr
from Products.CMFPlone.utils import safe_unicode
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.interfaces import IDXMeetingContent
from Products.PloneMeeting.utils import findMeetingAdvicePortalType
from Products.PloneMeeting.utils import get_event_field_data
from Products.PloneMeeting.utils import getAdvicePortalTypeIds
from Products.PloneMeeting.utils import getWorkflowAdapter
from Products.PloneMeeting.utils import historize_object_data
from Products.PloneMeeting.utils import isModifiedSinceLastVersion
from Products.PloneMeeting.utils import isPowerObserverForCfg
from Products.PloneMeeting.widgets.pm_richtext import PMRichTextFieldWidget
from z3c.form.browser.radio import RadioFieldWidget
from zope import schema
from zope.component import getAdapter
from zope.i18n import translate
from zope.interface import implements
from zope.schema.interfaces import IVocabularyFactory
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


class IMeetingAdvice(IDXMeetingContent):
    """
        MeetingAdvice schema
    """

    advice_group = LocalRoleField(
        title=_(u'title_advice_group'),
        description=_(u"Choose a group."),
        vocabulary=u'Products.PloneMeeting.content.advice.advice_group_vocabulary',
        required=True,
    )
    advice_type = schema.Choice(
        title=_(u'title_advice_type'),
        description=_(u"Choose an advice type."),
        vocabulary=u'Products.PloneMeeting.content.advice.advice_type_vocabulary',
        required=True,
    )
    form.widget('advice_hide_during_redaction', RadioFieldWidget)
    advice_hide_during_redaction = schema.Bool(
        title=_(u'title_advice_hide_during_redaction'),
        description=_("If you do not want the advice to be shown immediately after redaction, you can check this "
                      "box.  This will let you or other member of your group work on the advice before showing it.  "
                      "Note that if you lose access to the advice (for example if the item state evolve), "
                      "the advice will be considered 'Not given, was under edition'.  A manager will be able "
                      "to publish it nevertheless."),
        required=False,
    )
    form.widget('advice_comment', PMRichTextFieldWidget)
    advice_comment = RichText(
        title=_(u"title_advice_comment"),
        description=_("Enter the official comment."),
        required=False,
        allowed_mime_types=(u"text/html", )
    )
    form.widget('advice_observations', PMRichTextFieldWidget)
    advice_observations = RichText(
        title=_(u"title_advice_observations"),
        description=_("Enter optionnal observations if necessary."),
        required=False,
        allowed_mime_types=(u"text/html", )
    )
    advice_reference = schema.TextLine(
        title=_(u"title_advice_reference"),
        description=_("Enter a reference for this advice if necessary."),
        required=False,
    )
    form.mode(advice_row_id='hidden')
    advice_row_id = schema.TextLine(
        title=_(u"title_advice_row_id"),
        description=_("Linked advice row id, this is managed programmatically."),
        required=False,
    )


@form.default_value(field=IMeetingAdvice['advice_type'])
def advice_typeDefaultValue(data):
    res = ''
    tool = api.portal.get_tool('portal_plonemeeting')
    # check ToolPloneMeeting.advisersConfig
    advice_portal_type = findMeetingAdvicePortalType(data.context)
    for org_uid, adviser_infos in tool.adapted().get_extra_adviser_infos().items():
        if adviser_infos['portal_type'] == advice_portal_type:
            # use get in case overrided get_extra_adviser_infos and
            # 'default_advice_type' not managed, will be removable
            # when every profiles use new behavior
            res = adviser_infos.get('default_advice_type', '')
            break
    if not res:
        cfg = tool.getMeetingConfig(data.context)
        res = cfg and cfg.getDefaultAdviceType() or ''
    return res


@form.default_value(field=IMeetingAdvice['advice_hide_during_redaction'])
def advice_hide_during_redactionDefaultValue(data):
    published = data.context.REQUEST.get('PUBLISHED')
    if not published:
        return False
    tool = api.portal.get_tool('portal_plonemeeting')
    cfg = tool.getMeetingConfig(data.context)
    # manage when portal_type accessed from the Dexterity types configuration
    hidden = cfg and published.ti.id in cfg.getDefaultAdviceHiddenDuringRedaction() or False
    if hidden:
        api.portal.show_message(_("advice_hide_during_redaction_set_auto_to_true"),
                                request=data.context.REQUEST)
    return hidden


def get_advice_label(advice_info):
    """Render an advice label useable in several places."""
    res = advice_info["name"]
    if advice_info["delay"] and advice_info["delay_label"]:
        res = u"{0} - {1}".format(res, safe_unicode(advice_info["delay_label"]))
    return res


class MeetingAdvice(Container):
    """ """

    implements(IMeetingAdvice)
    # avoid inherited roles from the item or the item editor may edit the advice...
    __ac_local_roles_block__ = True

    security = ClassSecurityInfo()

    def getPrettyLink(self, **kwargs):
        """Return the IPrettyLink version of the title."""
        adapted = IPrettyLink(self)
        adapted.showContentIcon = True
        for k, v in kwargs.items():
            setattr(adapted, k, v)
        return adapted.getLink()

    def Title(self):
        '''
          This will construct the title of the advice, moreover, it checks for access
          to a confidential advice.
        '''
        # check that current user is not accessing to an advice that is confidential
        # to him but for which he knows the url to access to...
        parent = self.getParentNode()
        # in some case with plone.restapi summary serialize,
        # the parent is not found because self does not have acquisition
        if not parent:
            return ""
        if self.advice_group in parent.adviceIndex \
           and parent.adviceIndex[self.advice_group]['isConfidential']:
            tool = api.portal.get_tool('portal_plonemeeting')
            cfg = tool.getMeetingConfig(self)
            if not parent._adviceIsViewableForCurrentUser(
               cfg,
               isPowerObserverForCfg(cfg, cfg.getAdviceConfidentialFor()),
               parent.adviceIndex[self.advice_group]):
                raise Unauthorized

        # when creating a new advice object, it still not exist in parent's adviceIndex
        label = u""
        if self.advice_group in parent.adviceIndex:
            label = get_advice_label(parent.adviceIndex[self.advice_group])
        # we can not return a translated msg using _ so translate it
        return translate(
            "Advice ${advice_label} given on item ${item_title}",
            mapping={'item_title': unicode(parent.Title(), 'utf-8'),
                     'advice_label': label},
            domain="PloneMeeting",
            default="Advice given on item",
            context=self.REQUEST)

    def title_or_id(self):
        """ """
        return self.Title()

    def query_state(self):
        '''In what state am I ?'''
        wfTool = api.portal.get_tool('portal_workflow')
        return wfTool.getInfoFor(self, 'review_state')

    security.declarePublic('wfConditions')

    def wfConditions(self):
        '''Returns the adapter that implements the interface that proposes
           methods for use as conditions in the workflow associated with this
           item.'''
        return getWorkflowAdapter(self, conditions=True)

    security.declarePublic('wfActions')

    def wfActions(self):
        '''Returns the adapter that implements the interface that proposes
           methods for use as actions in the workflow associated with this
           item.'''
        return getWorkflowAdapter(self, conditions=False)

    def _updateAdviceRowId(self):
        '''Make sure advice_row_id is correct.'''
        # the row_id is stored in parent (item) adviceIndex
        item = self.getParentNode()

        # if a powerAdviser is adding an advice, the advice_group is not
        # in the item.adviceIndex, so if not found, check that
        if self.advice_group in item.adviceIndex:
            adviceInfo = item.adviceIndex[self.advice_group]
            row_id = adviceInfo['row_id']
        else:
            # check if it is actually a power adviser adding a not asked advice
            tool = api.portal.get_tool('portal_plonemeeting')
            cfg = tool.getMeetingConfig(item)
            if self.advice_group in cfg.getPowerAdvisersGroups():
                row_id = ''
            else:
                raise KeyError('Not able to find a value to set for advice row_id!')
        self.advice_row_id = row_id

    def _get_final_state_id(self):
        """By default final state is 'advice_given', but when using a custom WF,
           final state is the real final state in the WF."""
        wf_tool = api.portal.get_tool('portal_workflow')
        wf = wf_tool.getWorkflowsFor(self.portal_type)[0]
        final_state_ids = get_final_states(wf, ignored_transition_ids=['giveAdvice'])
        final_state_ids = len(final_state_ids) > 1 and \
            [state_id for state_id in final_state_ids if state_id != "advice_given"] or \
            final_state_ids
        final_state_id = final_state_ids[0]
        if wf.initial_state == final_state_id:
            return 'advice_given'
        else:
            return final_state_id

    def _get_final_transition_id(self):
        """Return the filal WF transition, useful when using a custom workflow,
           by default this will be the 'giveAdvice' transition."""
        wf_tool = api.portal.get_tool('portal_workflow')
        wf = wf_tool.getWorkflowsFor(self.portal_type)[0]
        return get_leading_transitions(wf, self._get_final_state_id())[0].id

    def get_advice_given_on(self):
        '''Return the date the advice was given on.
           If we do not use a custom workflow, returns the smallest date
           between modified() and last event 'giveAdvice'.
           This manages case when advice is edited after it is given, for example
           when a MeetingManager corrects a typo, the advice_given_on date will be
           the 'giveAdvice' date.'''
        final_transition_id = self._get_final_transition_id()
        lastEvent = getLastWFAction(self, final_transition_id)
        modified = self.modified()
        if not lastEvent:
            return modified
        else:
            # common case
            if final_transition_id == 'giveAdvice':
                return min(lastEvent['time'], modified)
            # custom advice WF with a real final state
            else:
                return lastEvent['time']

    def historize_if_relevant(self, comment):
        """Historize if self was never historized or
           if it was modified since last version."""
        # only historize advice if it was modified since last historization
        # and if it is not 'asked_again', indeed we do not historize an advice
        # that is 'asked_again' of it's predecessor would be an advice 'asked_again' too...
        if self.advice_type != 'asked_again' and isModifiedSinceLastVersion(self):
            historize_object_data(self, comment=comment)

    # def attribute_is_used_cachekey(method, self, name):
    #     '''cachekey method for self.attribute_is_used.'''
    #     return "{0}.{1}".format(self.portal_type, name)

    security.declarePublic('attribute_is_used')

    #  @ram.cache(attribute_is_used_cachekey)
    def attribute_is_used(self, name):
        '''Necessary for utils._addManagedPermissions for advice for now
           any attribute is used ?'''
        if name == 'advice_accounting_commitment':
            return base_hasattr(self, 'advice_accounting_commitment')
        return True

    def getIndexesRelatedTo(self, related_to='annex', check_deferred=True):
        '''See doc in interfaces.py.'''
        idxs = ['SearchableText']
        return idxs

    security.declarePublic('adapted')

    def adapted(self):
        '''Make adapted method available on advice, but actually no adapter
           can be defined, just return self.'''
        return self

    def get_previous_advice_type(self):
        """ """
        adapter = getAdapter(self, IImioHistory, 'advice_given')
        last_event = getLastAction(adapter)
        prev_advice_type = None
        if last_event:
            prev_advice_type = get_event_field_data(
                last_event["advice_data"], "advice_type")
        return prev_advice_type


class MeetingAdviceSchemaPolicy(DexteritySchemaPolicy):
    """ """

    def bases(self, schemaName, tree):
        return (IMeetingAdvice, )


class AdviceGroupVocabulary(object):
    implements(IVocabularyFactory)

    def __call__(self, context):
        """"""
        terms = []
        advicePortalTypeIds = getAdvicePortalTypeIds()

        # take into account groups for wich user can add an advice
        # while adding an advice, the context is his parent, aka a MeetingItem
        alterable_advice_org_uids = []
        if context.meta_type == 'MeetingItem':
            alterable_advice_org_uids = context.getAdvicesGroupsInfosForUser(compute_to_edit=False)[0]
        # take into account groups for which user can edit an advice
        elif context.portal_type in advicePortalTypeIds:
            alterable_advice_org_uids = context.getAdvicesGroupsInfosForUser(compute_to_add=False)[1]
            # make sure advice_group selected on advice is in the vocabulary
            if context.advice_group not in alterable_advice_org_uids:
                alterable_advice_org_uids.append(context.advice_group)

        # manage case where we have several meetingadvice portal_types
        # depending on current portal_type, clean up selectable orgs
        itemObj = context.meta_type == 'MeetingItem' and context or context.getParentNode()
        current_portal_type = findMeetingAdvicePortalType(context)
        alterable_advice_org_uids = [
            org_uid for org_uid in alterable_advice_org_uids
            if (itemObj.adapted()._advicePortalTypeForAdviser(org_uid) == current_portal_type or
                (context.portal_type in advicePortalTypeIds and org_uid == context.advice_group))]

        # create vocabulary
        for alterable_advice_org_uid in alterable_advice_org_uids:
            org = get_organization(alterable_advice_org_uid)
            terms.append(SimpleTerm(alterable_advice_org_uid,
                                    alterable_advice_org_uid,
                                    org.get_full_title()))
        return SimpleVocabulary(terms)


class AdviceTypeVocabulary(object):
    implements(IVocabularyFactory)

    def __call__(self, context):
        """ """
        terms = []
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(context)

        # manage when portal_type accessed from the Dexterity types configuration
        if cfg:
            # get usedAdviceTypes depending on current meetingadvice portal_type
            itemObj = context.meta_type == 'MeetingItem' and context or context.getParentNode()
            usedAdviceTypes = itemObj._adviceTypesForAdviser(
                findMeetingAdvicePortalType(context))

            # make sure if an adviceType was used for context and it is no more available, it
            # appears in the vocabulary and is so useable...
            if context.portal_type in getAdvicePortalTypeIds() and \
               context.advice_type not in usedAdviceTypes:
                usedAdviceTypes += (context.advice_type, )
            # build vocabulary terms
            for term in get_vocab(
                    tool,
                    'ConfigAdviceTypes',
                    include_asked_again=True,
                    include_term_id=False)._terms:
                if term.token in usedAdviceTypes:
                    terms.append(SimpleTerm(term.value, term.token, term.title))
        return SimpleVocabulary(terms)



# -*- coding: utf-8 -*-
#
# File: testMeetingConfig.py
#
# GNU General Public License (GPL)
#

from collective.compoundcriterion.interfaces import ICompoundCriterionFilter
from collective.eeafaceted.collectionwidget.utils import getCollectionLinkCriterion
from DateTime import DateTime
from datetime import datetime
from datetime import timedelta
from ftw.labels.interfaces import ILabeling
from imio.helpers.cache import cleanRamCacheFor
from imio.helpers.content import richtextval
from plone import api
from plone.app.querystring.querybuilder import queryparser
from plone.dexterity.utils import createContentInContainer
from Products.Archetypes.event import ObjectEditedEvent
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.permissions import View
from Products.PloneMeeting.adapters import _find_nothing_query
from Products.PloneMeeting.tests.PloneMeetingTestCase import PloneMeetingTestCase
from Products.PloneMeeting.tests.PloneMeetingTestCase import pm_logger
from Products.PloneMeeting.utils import getAdvicePortalTypes
from zope.component import getAdapter
from zope.component import getAdapters
from zope.event import notify


class testSearches(PloneMeetingTestCase):
    '''Tests the adapters used for searches.'''

    def test_pm_DefaultSelectedSearch(self):
        '''Test that the 'meetingAppDefaultView' defined in the MeetingConfig data
           is correctly applied to the collection-link widget on the facetednavigation view.'''
        # selected meetingAppDefaultView is 'searchallitems'
        cfg = self.meetingConfig
        searchallitems = cfg.searches.searches_items.searchallitems
        collectionCriterion = getCollectionLinkCriterion(cfg.searches)
        self.assertEqual(collectionCriterion.default, searchallitems.UID())

    def test_pm_SearchItemsOfMyGroups(self):
        '''Test the 'items-of-my-groups' adapter that returns items using proposingGroup
           the current user is in.'''
        cfg = self.meetingConfig
        itemTypeName = cfg.getItemTypeName()

        # siteadmin is not member of any PloneMeeting groups
        adapter = getAdapter(cfg, ICompoundCriterionFilter, name='items-of-my-groups')
        self.changeUser('siteadmin')
        self.assertEqual(adapter.query,
                         {'getProposingGroup': {'query': []},
                          'portal_type': {'query': itemTypeName}})

        # pmManager is member of 'developers' and 'vendors'
        self.changeUser('pmManager')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemsofmygroups')
        self.assertEqual(adapter.query,
                         {'getProposingGroup': {'query': [self.developers_uid, self.vendors_uid]},
                          'portal_type': {'query': itemTypeName}})

        # pmCreator1 is member of 'developers'
        self.changeUser('pmCreator1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemsofmygroups')
        self.assertEqual(adapter.query,
                         {'getProposingGroup': {'query': [self.developers_uid]},
                          'portal_type': {'query': itemTypeName}})

        # a deactivated group is still listed
        self.changeUser('siteadmin')
        self._select_organization(self.developers_uid, remove=True)
        self.changeUser('pmManager')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemsofmygroups')
        self.assertEqual(adapter.query,
                         {'getProposingGroup': {'query': [self.developers_uid, self.vendors_uid]},
                          'portal_type': {'query': itemTypeName}})

    def test_pm_SearchItemsToAdviceAdapter(self):
        '''Test the 'items-to-advice' adapter that should return a list of items
           a user has to give an advice for.'''
        self.changeUser('admin')
        cfg = self.meetingConfig
        cfg.setUsedAdviceTypes(cfg.getUsedAdviceTypes() + ('asked_again', ))
        cfg.setCustomAdvisers(
            [{'row_id': 'unique_id_123',
              'org': self.vendors_uid,
              'delay': '5', }, ])
        itemTypeName = cfg.getItemTypeName()

        # first test the generated query
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='items-to-advice')
        # admin is not adviser
        self.assertEqual(adapter.query,
                         {'indexAdvisers': {'query': []},
                          'portal_type': {'query': itemTypeName}})
        # as adviser, query is correct
        self.changeUser('pmAdviser1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstoadvice')
        self.assertEqual(
            adapter.query,
            {'indexAdvisers': {
                'query': ['{0}_advice_not_given'.format(self.developers_uid),
                          'delay__{0}_advice_not_given'.format(self.developers_uid),
                          '{0}_advice_asked_again'.format(self.developers_uid),
                          'delay__{0}_advice_asked_again'.format(self.developers_uid),
                          '{0}_advice_hidden_during_redaction'.format(self.developers_uid),
                          'delay__{0}_advice_hidden_during_redaction'.format(self.developers_uid)]},
                'portal_type': {'query': itemTypeName}})

        # now do the query
        # this adapter is used by the "searchallitemstoadvice"
        collection = cfg.searches.searches_items.searchallitemstoadvice
        # by default, no item to advice...
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstoadvice')
        self.failIf(collection.results())
        # an advice can be given when an item is 'proposed'
        self.assertEqual(cfg.getItemAdviceStates(), (self._stateMappingFor('proposed'), ))
        # create an item to advice
        self.changeUser('pmCreator1')
        item = self.create('MeetingItem')
        item.setOptionalAdvisers(
            (self.developers_uid, '{0}__rowid__unique_id_123'.format(self.vendors_uid)))
        item._update_after_edit()
        # as the item is "itemcreated", advices are not givable
        self.changeUser('pmAdviser1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstoadvice')
        self.failIf(collection.results())
        # now propose the item
        self.changeUser('pmCreator1')
        self.proposeItem(item)
        # only advisers can give an advice, so a creator for example will not see it
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstoadvice')
        self.failIf(collection.results())
        # now test as advisers
        self.changeUser('pmAdviser1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstoadvice')
        self.assertEqual(len(collection.results()), 1)
        self.assertEqual(collection.results()[0].UID, item.UID())
        # when an advice on an item is given, the item is no more returned by searchItemsToAdvice
        # so pmAdviser1 gives his advice
        createContentInContainer(item,
                                 'meetingadvice',
                                 **{'advice_group': self.developers_uid,
                                    'advice_type': u'positive',
                                    'advice_comment': richtextval(u'My comment')})
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstoadvice')
        self.failIf(collection.results())
        # pmReviewer2 is adviser for 'vendors', delay-aware advices are also returned
        self.changeUser('pmReviewer2')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstoadvice')
        self.assertEqual(len(collection.results()), 1)
        self.assertEqual(collection.results()[0].UID, item.UID())
        # when an advice on an item is given, the item is no more returned by searchItemsToAdvice
        # so pmReviewer2 gives his advice
        advice = createContentInContainer(item,
                                          'meetingadvice',
                                          **{'advice_group': self.vendors_uid,
                                             'advice_type': u'negative',
                                             'advice_comment': richtextval(u'My comment')})
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstoadvice')
        self.failIf(collection.results())

        # ask advice again, it will appear to 'pmReviewer2' in the query
        self.backToState(item, 'itemcreated')
        self.changeUser('pmCreator1')
        advice.restrictedTraverse('@@change-advice-asked-again')()
        self.proposeItem(item)
        self.changeUser('pmReviewer2')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstoadvice')
        self.assertEqual(len(collection.results()), 1)
        self.assertEqual(collection.results()[0].UID, item.UID())

        # a given advice that is 'hidden_during_redaction' is also found by this search
        advice.advice_type = u'positive'
        changeView = advice.restrictedTraverse('@@change-advice-hidden-during-redaction')
        changeView()
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstoadvice')
        self.assertEqual(len(collection.results()), 1)
        self.assertEqual(collection.results()[0].UID, item.UID())

    def test_pm_SearchItemsToAdviceWithoutHiddenDuringRedactionAdapter(self):
        '''Test the 'items-to-advice-without-hidden-during-redaction' adapter
           that should return a list of items a user has to give an advice for but not
           advice currently hidden during redaction.'''
        # just check query as full search is tested in test_pm_SearchItemsToAdviceAdapter
        # as adviser, query is correct
        self.changeUser('pmAdviser1')
        cfg = self.meetingConfig
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='items-to-advice-without-hidden-during-redaction')
        itemTypeName = cfg.getItemTypeName()
        self.assertEqual(
            adapter.query,
            {'indexAdvisers': {
                'query': ['{0}_advice_not_given'.format(self.developers_uid),
                          'delay__{0}_advice_not_given'.format(self.developers_uid),
                          '{0}_advice_asked_again'.format(self.developers_uid),
                          'delay__{0}_advice_asked_again'.format(self.developers_uid)]},
                'portal_type': {'query': itemTypeName}})

    def test_pm_SearchMyItemsToAdviceAdapter(self):
        '''Test the 'my-items-to-advice' adapter that should return
           a list of items a user has to give an advice for.
           This will return only items for which advice is asked to entire group or
           asked to current user.
           Advice asked to another user will not be returned.'''
        self.changeUser('admin')
        cfg = self.meetingConfig
        cfg.setUsedAdviceTypes(cfg.getUsedAdviceTypes() + ('asked_again', ))
        originalCustomAdvisers = {'row_id': 'unique_id_123',
                                  'org': self.vendors_uid,
                                  'gives_auto_advice_on': '',
                                  'for_item_created_from': '2012/01/01',
                                  'for_item_created_until': '',
                                  'gives_auto_advice_on_help_message': '',
                                  'delay': '10',
                                  'delay_left_alert': '',
                                  'delay_label': 'Delay label', }
        cfg.setCustomAdvisers([originalCustomAdvisers])
        cfg.setSelectableAdviserUsers((self.vendors_uid, ))
        cfg.setItemAdviceStates((self._stateMappingFor('itemcreated'), ))
        cfg.setItemAdviceEditStates((self._stateMappingFor('itemcreated'), ))
        cfg.setItemAdviceViewStates((self._stateMappingFor('itemcreated'), ))
        itemTypeName = cfg.getItemTypeName()

        # first test the generated query
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='my-items-to-advice')
        # admin is not adviser
        self.assertEqual(adapter.query,
                         {'indexAdvisers': {'query': []},
                          'portal_type': {'query': itemTypeName}})
        # as adviser, query is correct
        self.changeUser('pmReviewer2')
        self.assertEqual(
            adapter.query,
            {'indexAdvisers': {
                'query': [
                    '{0}_advice_not_given__userid__pmReviewer2'.format(
                        self.vendors_uid),
                    '{0}_advice_not_given__userid__entireadvisersgroup'.format(
                        self.vendors_uid),
                    'delay__{0}_advice_not_given__userid__pmReviewer2'.format(
                        self.vendors_uid),
                    'delay__{0}_advice_not_given__userid__entireadvisersgroup'.format(
                        self.vendors_uid),
                    '{0}_advice_asked_again__userid__pmReviewer2'.format(
                        self.vendors_uid),
                    '{0}_advice_asked_again__userid__entireadvisersgroup'.format(
                        self.vendors_uid),
                    'delay__{0}_advice_asked_again__userid__pmReviewer2'.format(
                        self.vendors_uid),
                    'delay__{0}_advice_asked_again__userid__entireadvisersgroup'.format(
                        self.vendors_uid),
                    '{0}_advice_hidden_during_redaction__userid__pmReviewer2'.format(
                        self.vendors_uid),
                    '{0}_advice_hidden_during_redaction__userid__entireadvisersgroup'.format(
                        self.vendors_uid),
                    'delay__{0}_advice_hidden_during_redaction__userid__pmReviewer2'.format(
                        self.vendors_uid),
                    'delay__{0}_advice_hidden_during_redaction__userid__entireadvisersgroup'.format(
                        self.vendors_uid)]},
                'portal_type': {'query': itemTypeName}})

        # now do the query
        # this adapter is used by the "searchmyitemstoadvice"
        myitemstoadvice = cfg.searches.searches_items.searchmyitemstoadvice
        itemstoadvice = cfg.searches.searches_items.searchallitemstoadvice
        # by default, no item to advice...
        self.failIf(myitemstoadvice.results())
        self.failIf(itemstoadvice.results())

        # create 3 items to advice
        # one with advice asked to entire vendors advisers
        # one with advice asked to pmAdviser1
        # one with advice asked to another vendors adviser "pmManager"
        self.changeUser('pmCreator1')
        self.create(
            'MeetingItem',
            optionalAdvisers=(self.vendors_uid, ))
        self.create(
            'MeetingItem',
            optionalAdvisers=('{0}__userid__pmReviewer2'.format(self.vendors_uid), ))
        self.create(
            'MeetingItem',
            optionalAdvisers=('{0}__userid__pmManager'.format(self.vendors_uid), ))

        # pmReviewer2 will only get item1/item2 in the "searchmyitemstoadvice" collection
        # the "searchallitemstoadvice" collection will return the 3 items
        self.changeUser('pmReviewer2')
        self.assertEqual(len(myitemstoadvice.results()), 2)
        self.assertEqual(len(itemstoadvice.results()), 3)

    def test_pm_SearchAdvisedItems(self):
        '''Test the 'advised-items' adapter.  This should return a list of items
           a user has already give an advice for.'''
        self.changeUser('admin')
        cfg = self.meetingConfig
        itemTypeName = cfg.getItemTypeName()
        cfg.setUsedAdviceTypes(cfg.getUsedAdviceTypes() + ('asked_again', ))

        # first test the generated query
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='advised-items')
        # admin is not adviser
        self.assertEqual(adapter.query,
                         {'indexAdvisers': {'query': []},
                          'portal_type': {'query': itemTypeName}})
        # as adviser, query is correct
        self.changeUser('pmAdviser1')
        indexAdvisers = []
        adviceStates = []
        for portal_type in getAdvicePortalTypes():
            adviceWF = self.wfTool.getWorkflowsFor(portal_type.id)[0]
            adviceStates += adviceWF.states.keys()
        # remove duplicates
        adviceStates = tuple(set(adviceStates))
        for adviceState in adviceStates:
            indexAdvisers.append('{0}_{1}'.format(self.developers_uid, adviceState))
            indexAdvisers.append('delay__{0}_{1}'.format(self.developers_uid, adviceState))

        cleanRamCacheFor('Products.PloneMeeting.adapters.query_adviseditems')
        self.assertEqual(adapter.query,
                         {'indexAdvisers': {'query': indexAdvisers},
                          'portal_type': {'query': itemTypeName}})

        # now do the query
        # this adapter is used by the "searchalladviseditems"
        collection = cfg.searches.searches_items.searchalladviseditems
        # by default, no advised item...
        self.changeUser('pmAdviser1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_adviseditems')
        self.failIf(collection.results())
        # an advice can be given when an item is 'proposed'
        self.assertEqual(cfg.getItemAdviceStates(), (self._stateMappingFor('proposed'), ))
        # create an item to advice
        self.changeUser('pmCreator1')
        item1 = self.create('MeetingItem')
        item1.setOptionalAdvisers((self.developers_uid,))
        self.proposeItem(item1)
        # give an advice
        self.changeUser('pmAdviser1')
        advice = createContentInContainer(item1,
                                          'meetingadvice',
                                          **{'advice_group': self.developers_uid,
                                             'advice_type': u'positive',
                                             'advice_comment': richtextval(u'My comment')})
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_adviseditems')
        self.failUnless(collection.results())
        # another user will not see given advices
        self.changeUser('pmCreator1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_adviseditems')
        self.failIf(collection.results())
        # other advisers of the same group will also see advised items
        self.changeUser('pmManager')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_adviseditems')
        self.failUnless(collection.results())
        # now create a second item and ask advice to the vendors (pmManager)
        # it will be returned for pmManager but not for pmAdviser1
        self.changeUser('pmCreator1')
        item2 = self.create('MeetingItem')
        item2.setOptionalAdvisers((self.vendors_uid,))
        self.proposeItem(item2)
        self.changeUser('pmManager')
        createContentInContainer(item2,
                                 'meetingadvice',
                                 **{'advice_group': self.vendors_uid,
                                    'advice_type': u'positive',
                                    'advice_comment': richtextval(u'My comment')})
        # pmManager will see 2 items and pmAdviser1, just one, none for a non adviser
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_adviseditems')
        self.failUnless(len(collection.results()) == 2)
        self.changeUser('pmAdviser1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_adviseditems')
        self.failUnless(len(collection.results()) == 1)
        self.changeUser('pmCreator1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_adviseditems')
        self.failIf(collection.results())

        # ask advice again to 'pmAdviser1'
        # if an advice is asked again, it stiil appears in the given advice
        # this way, if advice is asked again but no directly giveable, it is
        # still found in proposed search
        self.backToState(item1, 'itemcreated')
        advice.restrictedTraverse('@@change-advice-asked-again')()
        self.proposeItem(item1)
        self.changeUser('pmAdviser1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_adviseditems')
        self.failUnless(len(collection.results()) == 1)
        self.assertEqual(collection.results()[0].UID, item1.UID())

    def test_pm_SearchAdvisedItemsWithDelay(self):
        '''Test the 'advised-items-with-delay' adapter.  This should return a list
           of items a user has already give a delay-aware advice for.'''
        self.changeUser('admin')
        cfg = self.meetingConfig
        itemTypeName = cfg.getItemTypeName()

        # first test the generated query
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='advised-items-with-delay')
        # admin is not adviser
        self.assertEqual(adapter.query,
                         {'indexAdvisers': {'query': []},
                          'portal_type': {'query': itemTypeName}})
        # as adviser, query is correct
        self.changeUser('pmAdviser1')
        adviceStates = []
        for portal_type in getAdvicePortalTypes():
            adviceWF = self.wfTool.getWorkflowsFor(portal_type.id)[0]
            adviceStates += adviceWF.states.keys()
        # remove duplicates
        adviceStates = tuple(set(adviceStates))
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_adviseditemswithdelay')
        self.assertEqual(adapter.query,
                         {'indexAdvisers': {'query':
                          ['delay__{0}_{1}'.format(
                           self.developers_uid, adviceState) for adviceState in adviceStates]},
                          'portal_type': {'query': itemTypeName}})

        # now do the query
        # this adapter is used by the "searchalladviseditemswithdelay"
        collection = cfg.searches.searches_items.searchalladviseditemswithdelay
        # by default, no advised item...
        self.changeUser('pmAdviser1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_adviseditemswithdelay')
        self.failIf(collection.results())
        # an advice can be given when an item is 'proposed'
        self.assertEqual(cfg.getItemAdviceStates(), (self._stateMappingFor('proposed'), ))
        # create an item to advice
        self.changeUser('pmCreator1')
        item1 = self.create('MeetingItem')
        item1.setOptionalAdvisers((self.developers_uid,))
        self.proposeItem(item1)
        # give a non delay-aware advice
        self.changeUser('pmAdviser1')
        createContentInContainer(item1,
                                 'meetingadvice',
                                 **{'advice_group': self.developers_uid,
                                    'advice_type': u'positive',
                                    'advice_comment': richtextval(u'My comment')})
        # non delay-aware advices are not found
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_adviseditemswithdelay')
        self.failIf(collection.results())
        # now create a second item and ask a delay-aware advice
        self.changeUser('admin')
        originalCustomAdvisers = {'row_id': 'unique_id_123',
                                  'org': self.developers_uid,
                                  'gives_auto_advice_on': '',
                                  'for_item_created_from': '2012/01/01',
                                  'for_item_created_until': '',
                                  'gives_auto_advice_on_help_message': '',
                                  'delay': '10',
                                  'delay_left_alert': '',
                                  'delay_label': 'Delay label', }
        cfg.setCustomAdvisers([originalCustomAdvisers, ])
        self.changeUser('pmCreator1')
        item2 = self.create('MeetingItem')
        item2.setOptionalAdvisers(('{0}__rowid__unique_id_123'.format(self.developers_uid),))
        self.proposeItem(item2)
        self.changeUser('pmAdviser1')
        createContentInContainer(item2,
                                 'meetingadvice',
                                 **{'advice_group': self.developers_uid,
                                    'advice_type': u'positive',
                                    'advice_comment': richtextval(u'My comment')})
        # pmManager will see 2 items and pmAdviser1, just one, none for a non adviser
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_adviseditemswithdelay')
        self.failUnless(len(collection.results()) == 1)
        self.changeUser('pmCreator1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_adviseditemswithdelay')
        self.failIf(collection.results())

    def test_pm_SearchItemsInCopy(self):
        '''Test the 'items-in-copy' adapter.  This should return a list of items
           a user is in copy of.'''
        self.changeUser('admin')
        # specify that copyGroups can see the item when it is proposed
        cfg = self.meetingConfig
        self._enableField('copyGroups')
        cfg.setItemCopyGroupsStates((self._stateMappingFor('proposed'), 'validated', ))

        itemTypeName = cfg.getItemTypeName()

        # first test the generated query
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='items-in-copy')
        # admin does not belong to any group
        self.assertEqual(adapter.query,
                         {'getCopyGroups': {'query': []},
                          'portal_type': {'query': itemTypeName}})
        # as creator, query is correct
        self.changeUser('pmCreator1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemsincopy')
        self.assertEqual(adapter.query,
                         {'getCopyGroups': {
                          'query': sorted(['AuthenticatedUsers', self.developers_creators])},
                          'portal_type': {'query': itemTypeName}})

        # now do the query
        # this adapter is used by the "searchallitemsincopy"
        collection = cfg.searches.searches_items.searchallitemsincopy
        # create an item and set another proposing group in copy of
        item = self.create('MeetingItem')
        # give a view access to members of vendors, like pmReviewer2
        item.setCopyGroups((self.vendors_reviewers, ))
        item._update_after_edit()
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemsincopy')
        self.failIf(collection.results())
        # connect as a member of 'developers_reviewers'
        self.changeUser('pmReviewer2')
        # the item is not proposed so not listed
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemsincopy')
        self.failIf(collection.results())
        # propose the item, it will be listed
        self.proposeItem(item)
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemsincopy')
        self.failUnless(collection.results())

    def test_pm_SearchItemsInCopyWithAutoCopyGroups(self):
        '''Test the 'items-in-copy' adapter when using auto copyGroups.'''
        self.changeUser('admin')
        # specify that copyGroups can see the item when it is proposed
        cfg = self.meetingConfig
        self._enableField('copyGroups')
        cfg.setItemCopyGroupsStates((self._stateMappingFor('proposed'), 'validated', ))
        # configure an auto copyGroup, vendors_reviewers will be set
        # as auto copyGroup for every items
        self.vendors.as_copy_group_on = u"python: ['reviewers']"

        # this adapter is used by the "searchallitemsincopy"
        collection = cfg.searches.searches_items.searchallitemsincopy
        self.changeUser('pmCreator1')
        item = self.create('MeetingItem')
        self.assertEqual(item.getAllCopyGroups(auto_real_plone_group_ids=True),
                         (self.vendors_reviewers, ))
        self.proposeItem(item)
        self.changeUser('pmReviewer2')
        self.failUnless(collection.results())

    def test_pm_show_copy_groups_search(self):
        """Test MeetingConfig.show_copy_groups_search used to show items in copy searches."""
        cfg = self.meetingConfig
        self._enableField('copyGroups')
        self.assertEqual(cfg.getSelectableCopyGroups(),
                         (self.developers_reviewers, self.vendors_reviewers))
        self.changeUser('pmCreator1')
        self.assertFalse(cfg.show_copy_groups_search())
        self.changeUser('pmReviewer2')
        self.assertTrue(cfg.show_copy_groups_search())
        # not shown if copyGroups not used
        self._enableField('copyGroups', enable=False)
        self.changeUser('pmCreator1')
        self.assertFalse(cfg.show_copy_groups_search())
        self.changeUser('pmReviewer2')
        self.assertFalse(cfg.show_copy_groups_search())

    def test_pm_SearchMyItemsTakenOver(self):
        '''Test the 'my-items-taken-over' method.  This should return
           a list of items a user has taken over.'''
        self.changeUser('admin')
        # specify that copyGroups can see the item when it is proposed
        cfg = self.meetingConfig
        self._enableField('copyGroups')
        cfg.setItemCopyGroupsStates((self._stateMappingFor('proposed'), 'validated', ))

        itemTypeName = cfg.getItemTypeName()

        # first test the generated query
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='my-items-taken-over')
        # query is correct
        self.changeUser('pmManager')
        self.assertEqual(adapter.query,
                         {'portal_type': {'query': itemTypeName},
                          'getTakenOverBy': {'query': 'pmManager'}})

        # now do the query
        # this adapter is used by the "searchmyitemstakenover"
        collection = cfg.searches.searches_items.searchmyitemstakenover
        item = self.create('MeetingItem')
        # by default nothing is returned
        self.failIf(collection.results())
        # now take item over
        item.setTakenOverBy(self.member.getId())
        item.reindexObject(idxs=['getTakenOverBy', ])
        # now it is returned
        self.failUnless(collection.results())
        # takenOverBy is set back to '' on each transition
        self.proposeItem(item)
        self.assertTrue(not item.getTakenOverBy())
        self.failIf(collection.results())

        # query is not cached (this was the case before and there was a bug
        # because using forevercache and member_id changed
        self.changeUser('pmCreator1')
        self.assertEqual(adapter.query,
                         {'portal_type': {'query': itemTypeName},
                          'getTakenOverBy': {'query': 'pmCreator1'}})

    def _searchItemsToValidateOfHighestHierarchicLevelReviewerInfo(self, cfg):
        """ """
        return ['{0}__reviewprocess__{1}'.format(self.developers_uid,
                                                 self._stateMappingFor('proposed'))]

    def test_pm_SearchItemsToValidateOfHighestHierarchicLevel(self):
        '''Test the searchItemsToValidateOfHighestHierarchicLevel method.
           This should return a list of items a user ***really*** has to validate.
           Items to validate are items for which user is a reviewer and only regarding
           his higher hierarchic level.
           So a reviewer level 1 and level 2 will only see items in level 2, a reviewer in level
           1 (only) will only see items in level 1.'''
        self.changeUser('admin')
        cfg = self.meetingConfig
        itemTypeName = cfg.getItemTypeName()

        # first test the generated query
        self.changeUser('pmManager')
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='items-to-validate-of-highest-hierarchic-level')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofhighesthierarchiclevel')
        reviewProcessInfo = self._searchItemsToValidateOfHighestHierarchicLevelReviewerInfo(cfg)
        self.assertEqual(
            adapter.query,
            {'reviewProcessInfo':
                {'query': reviewProcessInfo},
             'portal_type': {'query': itemTypeName}})

        reviewers = cfg.reviewersFor()
        # activate 'prevalidation' if necessary
        if 'prereviewers' not in reviewers:
            self._enablePrevalidation(cfg)
        reviewers = cfg.reviewersFor()
        self.assertTrue('prereviewers' in reviewers)
        # now do the query
        # this adapter is used by the "searchitemstovalidate"
        collection = cfg.searches.searches_items.searchitemstovalidate
        # create an item
        self.changeUser('pmCreator1')
        item = self.create('MeetingItem')
        # jump to first level of validation
        self.do(item, self.TRANSITIONS_FOR_PROPOSING_ITEM_1[0])
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofhighesthierarchiclevel')
        self.failIf(collection.results())
        self.changeUser('pmReviewerLevel1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofhighesthierarchiclevel')
        self.failUnless(collection.results())
        # now as 'pmReviewerLevel2', the item should not be returned
        # as he only see items of his highest hierarchic level
        self.changeUser('pmReviewerLevel2')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofhighesthierarchiclevel')
        self.failIf(collection.results())
        # pass the item to second last level of hierarchy, where 'pmReviewerLevel2' is reviewer for
        self.changeUser('pmReviewerLevel1')
        # jump to last level of validation
        self.proposeItem(item)
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofhighesthierarchiclevel')
        self.failIf(collection.results())
        self.changeUser('pmReviewerLevel2')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofhighesthierarchiclevel')
        self.failUnless(collection.results())

        # now give a view on the item by 'pmReviewer2' and check if, as a reviewer,
        # the search does returns him the item, it should not as he is just a reviewer
        # but not able to really validate the new item
        self._enableField('copyGroups')
        review_states = reviewers[reviewers.keys()[0]]
        if 'prereviewers' in reviewers:
            review_states += ('prevalidated',)
        cfg.setItemCopyGroupsStates(review_states)
        item.setCopyGroups((self.vendors_reviewers, ))
        item._update_after_edit()
        self.changeUser('pmReviewer2')
        # the user can see the item
        self.failUnless(self.hasPermission(View, item))
        # but the search will not return it
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofhighesthierarchiclevel')
        self.failIf(collection.results())
        # if the item is validated, it will not appear for pmReviewer1 anymore
        self.changeUser('pmReviewer1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofhighesthierarchiclevel')
        self.failUnless(collection.results())
        self.validateItem(item)
        self.failIf(collection.results())

    def test_pm_SearchItemsToValidateOfHighestHierarchicLevelReturnsEveryLevels(self):
        '''When a user is developers_level3reviewers and vendors_level2reviewers,
           both groups must be queried.'''
        self.changeUser('siteadmin')
        cfg = self.meetingConfig
        # make sure we use default itemWFValidationLevels,
        # useful when test executed with custom profile
        self._setUpDefaultItemWFValidationLevels(cfg)
        self._enablePrevalidation(cfg)

        # make pmReviewer2 is vendors_prereviewers and developers_reviewers
        # add pmReviewer2 to developers_reviewers
        self.changeUser('pmReviewer2')
        member_groups = [grp_id for grp_id in self.member.getGroups()
                         if grp_id != 'AuthenticatedUsers']
        self._removePrincipalFromGroups('pmReviewer2', member_groups)
        self._addPrincipalToGroup('pmReviewer2', self.developers_reviewers)
        self._addPrincipalToGroup('pmReviewer2', self.vendors_prereviewers)
        self.assertItemsEqual(
            self.member.getGroups(),
            ['AuthenticatedUsers', self.developers_reviewers, self.vendors_prereviewers])

        # generated query
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='items-to-validate-of-highest-hierarchic-level')
        query = adapter.query
        self.assertEqual(len(query['reviewProcessInfo']['query']), 2)
        self.assertTrue('{0}__reviewprocess__prevalidated'.format(self.developers_uid)
                        in query['reviewProcessInfo']['query'])
        self.assertTrue('{0}__reviewprocess__proposed'.format(self.vendors_uid)
                        in query['reviewProcessInfo']['query'])

    def test_pm_SearchItemsToValidateOfMyReviewerGroups(self):
        '''Test the 'items-to-validate-of-my-reviewer-groups' adapter.
           This should return a list of items a user could validate at any level,
           so not only his highest hierarchic level.  This will return finally every items
           corresponding to Plone reviewer groups the user is in.'''
        cfg = self.meetingConfig
        self.changeUser('admin')

        # activate the 'pre_validation' wfAdaptation if it exists in current profile...
        # if not, then reviewers must be at least 2 elements long
        reviewers = cfg.reviewersFor()
        if not len(reviewers) > 1:
            self._enablePrevalidation(cfg)
        reviewers = cfg.reviewersFor()
        if not len(reviewers) > 1:
            pm_logger.info("Could not launch test 'test_pm_SearchItemsToValidateOfMyReviewerGroups' "
                           "because we need at least 2 levels of item validation.")

        # first test the generated query
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='items-to-validate-of-my-reviewer-groups')
        # if user si not a reviewer, we want the search to return
        # nothing so the query uses an unknown review_state
        itemTypeName = cfg.getItemTypeName()
        self.assertEqual(adapter.query, _find_nothing_query(itemTypeName))
        # for a reviewer, query is correct
        self.changeUser('pmReviewer1')
        # only reviewer for highest level
        reviewers = cfg.reviewersFor()
        self._removeUsersFromEveryGroups(['pmReviewer1'])
        self._addPrincipalToGroup('pmReviewer1',
                                  "{0}_{1}".format(self.developers_uid, reviewers.keys()[0]))
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofmyreviewergroups')
        query = adapter.query
        query['reviewProcessInfo']['query'].sort()
        states = reviewers.values()[0]
        self.assertEqual({'portal_type': {'query': itemTypeName},
                          'reviewProcessInfo': {
                          'query': sorted(['{0}__reviewprocess__{1}'.format(self.developers_uid, state)
                                           for state in states])}},
                         adapter.query)

        # now do the query
        # this adapter is not used by default, but is intended to be used with
        # the "searchitemstovalidate" collection so use it with it
        collection = cfg.searches.searches_items.searchitemstovalidate
        patchedQuery = list(collection.query)
        patchedQuery[0]['v'] = 'items-to-validate-of-my-reviewer-groups'
        collection.query = patchedQuery

        # create 2 items
        self.changeUser('pmCreator1')
        item1 = self.create('MeetingItem')
        item2 = self.create('MeetingItem')
        self.do(item1, self.TRANSITIONS_FOR_PROPOSING_ITEM_1[0])
        self.do(item2, self.TRANSITIONS_FOR_PROPOSING_ITEM_1[0])
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofmyreviewergroups')
        self.failIf(collection.results())
        # as first level user, he will see items
        self.changeUser('pmReviewerLevel1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofmyreviewergroups')
        self.assertEqual(collection.results().length, 2)
        # as second level user, he will not see items of first level also
        self.changeUser('pmReviewerLevel2')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofmyreviewergroups')
        self.failIf(collection.results())

        # define 'pmReviewerLevel2' as a prereviewer (first validation level reviewer)
        self._turnUserIntoPrereviewer(self.member)
        # change again to 'pmReviewerLevel2' so changes in his groups are taken into account
        self.changeUser('pmReviewerLevel2')
        # he can access first validation level items
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofmyreviewergroups')
        self.assertEqual(collection.results().length, 2)
        # move item1 to last validation level
        self.proposeItem(item1)
        # both items still returned by the search for 'pmReviewerLevel2'
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofmyreviewergroups')
        self.assertEqual(collection.results().length, 2)
        # but now, the search only returns item2 to 'pmReviewerLevel1'
        self.changeUser('pmReviewerLevel1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofmyreviewergroups')
        self.assertEqual(collection.results().length, 1)
        self.failUnless(collection.results()[0].UID == item2.UID())

    def runSearchItemsToValidateOfEveryReviewerLevelsAndLowerLevelsTest(self):
        '''
          Helper method for activating the test_pm_SearchItemsToValidateOfEveryReviewerLevelsAndLowerLevels
          test when called from a subplugin.
        '''
        return False

    def _searchItemsToValidateOfEveryReviewerLevelsAndLowerLevelsReviewerInfo(self, cfg):
        """ """
        reviewers = cfg.reviewersFor()
        reviewer_states = reviewers[cfg._highestReviewerLevel(self.member.getGroups())]
        return ['{0}__reviewprocess__{1}'.format(self.developers_uid, reviewer_state)
                for reviewer_state in reviewer_states]

    def test_pm_SearchItemsToValidateOfEveryReviewerLevelsAndLowerLevels(self):
        '''Test the searchItemsToValidateOfEveryReviewerLevelsAndLowerLevels method.
           This will return items to validate of his highest hierarchic level and every levels
           under, even if user is not in the corresponding Plone reviewer groups.'''
        cfg = self.meetingConfig
        # check if self.runSearchItemsToValidateOfEveryReviewerLevelsAndLowerLevelsTest() is True
        if not self.runSearchItemsToValidateOfEveryReviewerLevelsAndLowerLevelsTest():
            pm_logger.info(
                "Test 'test_pm_SearchItemsToValidateOfEveryReviewerLevelsAndLowerLevels' was bypassed.")
            return
        self._enablePrevalidation(cfg, enable_extra_suffixes=True)
        itemTypeName = cfg.getItemTypeName()
        # create 2 items
        self.changeUser('pmCreator1')
        item1 = self.create('MeetingItem')
        item2 = self.create('MeetingItem')
        self.do(item1, self.TRANSITIONS_FOR_PROPOSING_ITEM_1[0])
        self.do(item2, self.TRANSITIONS_FOR_PROPOSING_ITEM_1[0])
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='items-to-validate-of-every-reviewer-levels-and-lower-levels')
        self.changeUser('pmObserver1')
        self.assertEqual(adapter.query, _find_nothing_query(itemTypeName))
        # now do the query
        # this adapter is not used by default, but is intended to be used with
        # the "searchitemstovalidate" collection so use it with it
        collection = cfg.searches.searches_items.searchitemstovalidate
        patchedQuery = list(collection.query)
        patchedQuery[0]['v'] = 'items-to-validate-of-every-reviewer-levels-and-lower-levels'
        collection.query = patchedQuery
        self.failIf(collection.results())
        # as first level user, he will see items
        self.changeUser('pmReviewerLevel1')
        # find state to use for current reviewer
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofeveryreviewerlevelsandlowerlevels')
        reviewProcessInfo = self._searchItemsToValidateOfEveryReviewerLevelsAndLowerLevelsReviewerInfo(cfg)
        self.assertEqual(adapter.query,
                         {'portal_type': {'query': itemTypeName},
                          'reviewProcessInfo': {'query': reviewProcessInfo}})
        self.failUnless(len(collection.results()) == 2)
        # as second level user, he will also see items because items are from lower reviewer levels
        self.changeUser('pmReviewerLevel2')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofeveryreviewerlevelsandlowerlevels')
        self.failUnless(len(collection.results()) == 2)

        # now propose item1, both items are still viewable to 'pmReviewerLevel2', but 'pmReviewerLevel1'
        # will only see item of 'his' highest hierarchic level
        self.proposeItem(item1)
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofeveryreviewerlevelsandlowerlevels')
        self.failUnless(len(collection.results()) == 2)
        self.changeUser('pmReviewerLevel1')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstovalidateofeveryreviewerlevelsandlowerlevels')
        self.failUnless(len(collection.results()) == 1)
        self.failUnless(collection.results()[0].UID == item2.UID())

    def test_pm_SearchItemsToCorrect(self):
        '''Test the 'items-to-correct' CompoundCriterion adapter.  This should return
           a list of items in state 'returned_to_proposing_group' the current user is able to edit.'''
        cfg = self.meetingConfig
        if not self._check_wfa_available(['return_to_proposing_group']):
            pm_logger.info("Bypassing test test_pm_SearchItemsToCorrect because it "
                           "needs the 'return_to_proposing_group' wfAdaptation.")
            return

        itemTypeName = cfg.getItemTypeName()
        self.changeUser('siteadmin')
        # first test the generated query
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='items-to-correct')
        # wfAdaptation 'return_to_proposing_group' is not enabled
        self.assertEqual(adapter.query, _find_nothing_query(itemTypeName))
        wfAdaptations = list(cfg.getWorkflowAdaptations())
        if 'return_to_proposing_group' not in wfAdaptations:
            wfAdaptations.append('return_to_proposing_group')
        cfg.setWorkflowAdaptations(wfAdaptations)
        notify(ObjectEditedEvent(cfg))

        # normally this search is not available to users that are not able to correct items
        # nevertheless, if a user is not able to edit items to correct, the special
        # query 'return nothing' is returned
        self.assertEqual(adapter.query, _find_nothing_query(itemTypeName))
        self.changeUser('pmManager')
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstocorrect')
        self.assertEqual(
            adapter.query,
            {'portal_type': {'query': itemTypeName},
             'reviewProcessInfo': {
             'query': ['{0}__reviewprocess__returned_to_proposing_group'.format(self.developers_uid)]}})

        # it returns only items the current user is able to correct
        # create an item for developers and one for vendors and 'return' it to proposingGroup
        self.create('Meeting')
        developersItem = self.create('MeetingItem')
        self.assertEqual(developersItem.getProposingGroup(), self.developers_uid)
        self.presentItem(developersItem)
        self.changeUser('pmCreator2')
        vendorsItem = self.create('MeetingItem')
        self.assertEqual(vendorsItem.getProposingGroup(), self.vendors_uid)
        self.changeUser('pmManager')
        self.presentItem(vendorsItem)
        collection = cfg.searches.searches_items.searchitemstocorrect
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstocorrect')
        self.failIf(collection.results())
        self.do(developersItem, 'return_to_proposing_group')
        self.do(vendorsItem, 'return_to_proposing_group')

        # pmManager may only edit developersItem
        self.assertTrue(self.hasPermission(ModifyPortalContent, developersItem))
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstocorrect')
        res = collection.results()
        self.failUnless(len(res) == 1)
        self.failUnless(res[0].UID == developersItem.UID())

        # pmCreator2 may only edit vendorsItem
        self.changeUser('pmCreator2')
        self.assertTrue(self.hasPermission(ModifyPortalContent, vendorsItem))
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstocorrect')
        res = collection.results()
        self.failUnless(len(res) == 1)
        self.failUnless(res[0].UID == vendorsItem.UID())

        # when an item is corrected, it appears in the correcteditems search
        collection = cfg.searches.searches_items.searchcorrecteditems
        self.assertEqual(len(collection.results()), 0)
        self.do(vendorsItem, 'backTo_presented_from_returned_to_proposing_group')
        self.assertEqual(len(collection.results()), 1)

    def test_pm_SearchItemsToCorrectToValidateOfHighestHierarchicLevel(self):
        '''Test the 'items-to-correct-to-validate-of-highest-hierarchic-level'
           CompoundCriterion adapter. This should return a list of items in state
           'returned_to_proposing_group_proposed' the current user is able to edit.'''
        cfg = self.meetingConfig
        if not self._check_wfa_available(['return_to_proposing_group_with_last_validation']):
            pm_logger.info(
                "Bypassing test test_pm_SearchItemsToCorrectToValidateHighestHierarchicLevel because it "
                "needs the 'return_to_proposing_group_with_last_validation' wfAdaptation.")
            return

        itemTypeName = cfg.getItemTypeName()
        self.changeUser('siteadmin')
        wfAdaptations = list(cfg.getWorkflowAdaptations())
        if 'return_to_proposing_group_with_last_validation' not in wfAdaptations:
            wfAdaptations.append('return_to_proposing_group_with_last_validation')
        # desactivate simple return to proposing group wf
        if 'return_to_proposing_group' in wfAdaptations:
            wfAdaptations.remove('return_to_proposing_group')
        cfg.setWorkflowAdaptations(wfAdaptations)
        notify(ObjectEditedEvent(cfg))

        # first test the generated query
        self.changeUser('pmManager')
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='items-to-correct-to-validate-of-highest-hierarchic-level')
        self.assertEqual(adapter.query, {
            'reviewProcessInfo':
            {'query': ['{0}__reviewprocess__returned_to_proposing_group_{1}'.format(
                self.developers_uid, self._stateMappingFor('proposed'))]},
            'portal_type': {'query': itemTypeName}})

        # it returns only items the current user is able to correct
        # create an item for developers and one for vendors and 'return' it to proposingGroup
        self.create('Meeting')
        developersItem = self.create('MeetingItem')
        self.assertEqual(developersItem.getProposingGroup(), self.developers_uid)
        self.presentItem(developersItem)
        self.changeUser('pmCreator2')
        vendorsItem = self.create('MeetingItem')
        self.assertEqual(vendorsItem.getProposingGroup(), self.vendors_uid)
        self.changeUser('pmManager')
        self.presentItem(vendorsItem)
        collection = cfg.searches.searches_items.searchitemstocorrecttovalidate
        cleanRamCacheFor('Products.PloneMeeting.adapters.query_itemstocorrecttovalidateofhighesthierarchiclevel')
        self.failIf(collection.results())
        self.do(developersItem, 'return_to_proposing_group')
        self.do(vendorsItem, 'return_to_proposing_group')

        self.changeUser('pmCreator1')
        self.do(developersItem, 'goTo_returned_to_proposing_group_' + self._stateMappingFor('proposed'))
        self.changeUser('pmCreator2')
        self.do(vendorsItem, 'goTo_returned_to_proposing_group_' + self._stateMappingFor('proposed'))
        self.changeUser('pmManager')

        # pmManager may only edit developersItem
        self.assertTrue(self.hasPermission(ModifyPortalContent, developersItem))
        cleanRamCacheFor(
            'Products.PloneMeeting.adapters.query_itemstocorrecttovalidateofhighesthierarchiclevel')
        res = collection.results()
        self.failUnless(len(res) == 1)
        self.failUnless(res[0].UID == developersItem.UID())

        # pmCreator2 can't edit vendorsItem
        self.changeUser('pmCreator2')
        self.assertTrue(not self.hasPermission(ModifyPortalContent, vendorsItem))
        cleanRamCacheFor(
            'Products.PloneMeeting.adapters.query_itemstocorrecttovalidateofhighesthierarchiclevel')
        res = collection.results()
        self.failUnless(len(res) == 0)

        # pmReviewer2 may only edit vendorsItem
        self.changeUser('pmReviewer2')
        self.assertTrue(self.hasPermission(ModifyPortalContent, vendorsItem))
        cleanRamCacheFor(
            'Products.PloneMeeting.adapters.query_itemstocorrecttovalidateofhighesthierarchiclevel')
        res = collection.results()
        self.failUnless(len(res) == 1)
        self.failUnless(res[0].UID == vendorsItem.UID())

    def test_pm_SearchAllItemsToValidateOfHighestHierarchicLevel(self):
        '''Test the 'all-items-to-validate-of-highest-hierarchic-level'
           CompoundCriterion adapter. This should return every items the user is able to validate
           so items that are 'proposed' and items that are 'returned_to_proposing_group_proposed'.'''
        # specify that copyGroups can see the item when it is proposed
        cfg = self.meetingConfig
        if not self._check_wfa_available(['return_to_proposing_group_with_last_validation']):
            pm_logger.info(
                "Bypassing test test_pm_SearchAllItemsToValidateHighestHierarchicLevel because it "
                "needs the 'return_to_proposing_group_with_last_validation' wfAdaptation.")
            return
        itemTypeName = cfg.getItemTypeName()
        self.changeUser('pmManager')
        # first test the generated query
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='all-items-to-validate-of-highest-hierarchic-level')
        states = self._get_query_review_process(cfg)[-1:]
        query = sorted(['{0}__reviewprocess__{1}'.format(self.developers_uid, state)
                        for state in states])
        query += sorted(['{0}__reviewprocess__returned_to_proposing_group_{1}'.format(self.developers_uid, state)
                         for state in states])
        self.assertEqual({'portal_type': {'query': [itemTypeName]},
                          'reviewProcessInfo': {'query': query}},
                         adapter.query)

    def _test_reviewer_groups(self, developersItem, vendorsItem, collection):
        self.changeUser('pmCreator1')
        self.do(developersItem,
                'goTo_returned_to_proposing_group_{}'.format(self._stateMappingFor('proposed_first_level')))
        self.changeUser('pmCreator2')
        self.do(vendorsItem,
                'goTo_returned_to_proposing_group_{}'.format(self._stateMappingFor('proposed_first_level')))

        self.changeUser('pmCreator1')
        # pmManager can't edit developersItem
        self.assertFalse(self.hasPermission(ModifyPortalContent, developersItem))
        cleanRamCacheFor(
            'Products.PloneMeeting.adapters.query_itemstocorrecttovalidateofeveryreviewerlevelsandlowerlevels')
        self.assertEqual(collection.results().length, 0)

        # pmCreator2 can't edit vendorsItem
        self.changeUser('pmCreator2')
        self.assertFalse(self.hasPermission(ModifyPortalContent, vendorsItem))
        cleanRamCacheFor(
            'Products.PloneMeeting.adapters.query_itemstocorrecttovalidateofeveryreviewerlevelsandlowerlevels')
        self.assertEqual(collection.results().length, 0)

        self.changeUser('admin')
        self.do(developersItem, 'goTo_returned_to_proposing_group_prevalidated')
        self.do(vendorsItem, 'goTo_returned_to_proposing_group_prevalidated')

        # pmReviewer1 may only edit developersItem
        self.changeUser('pmReviewer1')
        self.assertTrue(self.hasPermission(ModifyPortalContent, developersItem))
        cleanRamCacheFor(
            'Products.PloneMeeting.adapters.query_itemstocorrecttovalidateofeveryreviewerlevelsandlowerlevels')
        res = collection.results()
        self.assertEqual(res.length, 1)
        self.assertEqual(res[0].UID, developersItem.UID())

        # pmReviewer2 may only edit vendorsItem
        self.changeUser('pmReviewer2')
        self.assertTrue(self.hasPermission(ModifyPortalContent, vendorsItem))
        cleanRamCacheFor(
            'Products.PloneMeeting.adapters.query_itemstocorrecttovalidateofeveryreviewerlevelsandlowerlevels')
        res = collection.results()
        self.assertEqual(res.length, 1)
        self.assertEqual(res[0].UID, vendorsItem.UID())

    def test_pm_SearchItemsToCorrectToValidateOfEveryReviewerGroups(self):
        '''Test the 'items-to-correct-to-validate-of-every-reviewer-groups'
           CompoundCriterion adapter.  This should return a list of items in state
           'returned_to_proposing_group_proposed' the current user is able to edit.'''
        if not self._check_wfa_available(['return_to_proposing_group_with_all_validations']):
            pm_logger.info(
                "Bypassing test test_pm_SearchItemsToCorrectToValidateOfEveryReviewerGroups because it "
                "needs the 'return_to_proposing_group_with_all_validations' wfAdaptation.")
            return

        cfg = self.meetingConfig
        itemTypeName = cfg.getItemTypeName()

        self.changeUser('siteadmin')
        # first test the generated query
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='items-to-correct-to-validate-of-every-reviewer-groups')
        # wfAdaptation 'return_to_proposing_group_with_last_validation' is not enabled
        self.assertEqual(adapter.query, _find_nothing_query(itemTypeName))
        wfAdaptations = list(cfg.getWorkflowAdaptations())
        if 'return_to_proposing_group_with_all_validations' not in wfAdaptations:
            wfAdaptations.append('return_to_proposing_group_with_all_validations')
        # deactivate simple return to proposing group wf
        if 'return_to_proposing_group' in wfAdaptations:
            wfAdaptations.remove('return_to_proposing_group')
        cfg.setWorkflowAdaptations(wfAdaptations)
        self._enablePrevalidation(cfg)
        notify(ObjectEditedEvent(cfg))

        # normally this search is not available to users that are not able to review items
        # nevertheless, if a user is in not able to edit items to correct in proposed, the special
        # query 'return nothing' is returned
        self.assertEqual(adapter.query, _find_nothing_query(itemTypeName))
        self.changeUser('pmManager')
        cleanRamCacheFor(
            'Products.PloneMeeting.adapters.query_itemstocorrecttovalidateofeveryreviewerlevelsandlowerlevels')
        states = self._get_query_review_process(cfg)[1:]
        query = sorted(['{0}__reviewprocess__returned_to_proposing_group_{1}'.format(
                        self.developers_uid, state) for state in states])
        self.assertEqual({
            'portal_type': {'query': itemTypeName},
            'reviewProcessInfo': {'query': query}},
            adapter.query)

        # it returns only items the current user is able to correct
        # create an item for developers and one for vendors and 'return' it to proposingGroup
        self.create('Meeting')
        developers_item = self.create('MeetingItem')
        self.assertEqual(developers_item.getProposingGroup(), self.developers_uid)
        self.changeUser('pmCreator2')
        vendors_item = self.create('MeetingItem')
        self.assertEqual(vendors_item.getProposingGroup(), self.vendors_uid)
        # present items
        self.changeUser('pmManager')
        self.presentItem(developers_item)
        self.presentItem(vendors_item)
        collection = cfg.searches.searches_items.searchitemstocorrecttovalidateoffeveryreviewergroups
        cleanRamCacheFor(
            'Products.PloneMeeting.adapters.query_itemstocorrecttovalidateofeveryreviewerlevelsandlowerlevels')
        self.failIf(collection.results())

        self.do(developers_item, 'return_to_proposing_group')
        self.do(vendors_item, 'return_to_proposing_group')

        self._test_reviewer_groups(developers_item, vendors_item, collection)

    def _get_query_review_process(self, cfg):
        return [state['state'] for state in cfg.getItemWFValidationLevels()
                if state['enabled'] == '1' and state['state']]

    def test_pm_SearchAllItemsToValidateOfEveryReviewerGroups(self):
        '''Test the 'all-items-to-validate-of-every-reviewer-groups'
           CompoundCriterion adapter. This should return every items the user is able to validate
           so items that are 'proposed' and items that are 'returned_to_proposing_group_proposed'.'''
        cfg = self.meetingConfig
        if not self._check_wfa_available(['return_to_proposing_group_with_all_validations']):
            pm_logger.info(
                "Bypassing test test_pm_SearchAllItemsToValidateOfEveryReviewerGroups because it "
                "needs the 'return_to_proposing_group_with_all_validations' wfAdaptation.")
            return

        itemTypeName = cfg.getItemTypeName()
        self._enablePrevalidation(cfg)

        self.changeUser('pmManager')
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='all-items-to-validate-of-every-reviewer-groups')
        states = self._get_query_review_process(cfg)[1:]
        query = sorted(['{0}__reviewprocess__{1}'.format(self.developers_uid, state)
                        for state in states])
        query += sorted(['{0}__reviewprocess__returned_to_proposing_group_{1}'.format(self.developers_uid, state)
                         for state in states])
        self.assertEqual({'portal_type': {'query': [itemTypeName]},
                          'reviewProcessInfo': {'query': query}},
                         adapter.query)

    def test_pm_SearchUnreadItems(self):
        '''Test the 'items-with-negative-personal-labels' adapter.
           This should return a list of items for which current user did not checked the 'lu' label.'''
        # disable every showNumberOfItems so no search is enabled by custom profile
        cfg = self.meetingConfig
        for collection in cfg.searches.searches_items.objectValues():
            collection.showNumberOfItems = False
        cfg = self.meetingConfig
        self._enableField('labels')
        collection = cfg.searches.searches_items.searchunreaditems
        collection_uid = collection.UID()
        collection.showNumberOfItems = True

        # create item, not 'lu' by default
        self.changeUser('pmCreator1')
        # check that counter is updated when a personal label is changed
        json_collections_count = self.getMeetingFolder().restrictedTraverse("@@json_collections_count")
        self.assertEqual(
            json_collections_count(),
            '{"criterionId": "c1", "countByCollection": [{"count": 0, "uid": "%s"}]}' % collection_uid)
        item = self.create('MeetingItem')
        # for now item is not 'lu'
        self.assertEqual(len(collection.results()), 1)
        self.assertEqual(
            json_collections_count(),
            '{"criterionId": "c1", "countByCollection": [{"count": 1, "uid": "%s"}]}' % collection_uid)
        # make item 'lu'
        labeling = ILabeling(item)
        labeling.pers_update(['lu'], True)
        item.reindexObject(idxs=['labels'])
        self.assertEqual(len(collection.results()), 0)
        self.assertEqual(
            json_collections_count(),
            '{"criterionId": "c1", "countByCollection": [{"count": 0, "uid": "%s"}]}' % collection_uid)

    def test_pm_CompoundCriterionAdapterItemsWithNegativePreviousIndex(self):
        '''Test the 'items-with-negative-previous-index' adapter.
           Here we will try to get items that do not have a certain advice asked.'''
        cfg = self.meetingConfig
        cfg.setCustomAdvisers(
            [{'row_id': 'unique_id_123',
              'org': self.developers_uid,
              'delay': '10', }, ])
        # this will return items for which developers 10 days delay advice was not asked
        # or vendors advice was not asked
        collection = cfg.searches.searches_items.searchallitems
        query = collection.query
        query.append({u'i': u'indexAdvisers',
                      u'o': u'plone.app.querystring.operation.selection.is',
                      u'v': [u'delay_row_id__unique_id_123',
                             u'real_org_uid__{0}'.format(self.vendors_uid)]})
        query.append({u'i': u'CompoundCriterion',
                      u'o': u'plone.app.querystring.operation.compound.is',
                      u'v': [u'items-with-negative-previous-index']})
        collection.setQuery(query)
        self.assertEqual(
            queryparser.parseFormquery(collection, collection.getQuery())[u'indexAdvisers'],
            {'not': [u'delay_row_id__unique_id_123',
                     u'real_org_uid__{0}'.format(self.vendors_uid)]})

        # test
        self.changeUser('pmCreator1')
        self.assertEqual(len(collection.results()), 0)
        item = self.create('MeetingItem')
        self.assertEqual(len(collection.results()), 1)
        # ask advices
        item.setOptionalAdvisers(('{0}__rowid__unique_id_123'.format(self.developers_uid), ))
        item._update_after_edit()
        item.reindexObject(idxs=['indexAdvisers'])
        self.assertEqual(len(collection.results()), 0)
        item.setOptionalAdvisers(())
        item._update_after_edit()
        item.reindexObject(idxs=['indexAdvisers'])
        self.assertEqual(len(collection.results()), 1)
        item.setOptionalAdvisers((self.vendors_uid, ))
        item._update_after_edit()
        item.reindexObject(idxs=['indexAdvisers'])
        self.assertEqual(len(collection.results()), 0)

    def test_pm_DashboardCollectionsAreEditable(self):
        """This will ensure created DashboardCollections are editable.
           It could not be the case when using a wrong query."""
        self.changeUser('siteadmin')
        cfg = self.meetingConfig
        brains = api.content.find(context=cfg.searches, portal_type='DashboardCollection')
        for brain in brains:
            collection = brain.getObject()
            self.assertTrue(collection.restrictedTraverse('edit')())

    def test_pm_SearchLastDecisions(self):
        '''Test the 'last-decisions' CompoundCriterion adapter.
           This should decided meetings from 60 days in the past to 60 days in the future.'''
        cfg = self.meetingConfig
        meetingTypeName = cfg.getMeetingTypeName()

        # siteadmin is not member of any PloneMeeting groups
        collection = cfg.searches.searches_decisions.searchlastdecisions

        adapter = getAdapter(collection, ICompoundCriterionFilter, name='last-decisions')
        self.changeUser('siteadmin')
        # meeting_date minmax is correct, first date is 60 days before and second 60 days after now
        self.assertTrue(adapter.query['meeting_date']['query'][0] < DateTime() - 59)
        self.assertTrue(adapter.query['meeting_date']['query'][1] > DateTime() + 59)
        self.assertEqual(adapter.query['portal_type']['query'], [meetingTypeName])

        # decided meetings in the future and in the past are found
        self.changeUser('pmManager')
        self.failIf(collection.results())
        now = datetime.now()
        past_meeting = self.create('Meeting', date=now - timedelta(days=45))
        future_meeting = self.create('Meeting', date=now + timedelta(days=45))
        self.failIf(collection.results())
        self.decideMeeting(past_meeting)
        self.decideMeeting(future_meeting)
        result_uids = [brain.UID for brain in collection.results()]
        self.assertTrue(past_meeting.UID() in result_uids)
        self.assertTrue(future_meeting.UID() in result_uids)

    def test_pm_EverySearchesUseDifferentCachedMethod(self):
        """Make sure a different method is used for caching because every adapters
           use the "query" method but ram.cache would have one single cache entry
           because it's key is module path + method name so
           Products.PloneMeeting.adapters.query."""
        cfg = self.meetingConfig
        adapters = getAdapters([cfg], ICompoundCriterionFilter)
        query_aliases = []
        for adapter_name, adapter_instance in adapters:
            query_methods = [method_name for method_name in dir(adapter_instance)
                             if method_name.startswith('query_')]
            # there must be at least one query_... method that is an alias for query
            self.assertTrue(
                query_methods,
                "No query_xxx methdod for {0}".format(adapter_name))
            # make sure the query_... method is an alias for query
            found_alias = False
            for query_method in query_methods:
                adapter_class = adapter_instance.__class__
                if getattr(adapter_class, query_method) == adapter_class.query:
                    found_alias = True
                    break
            self.assertTrue(found_alias,
                            "Alias not found for {0}".format(adapter_name))
            # keep the query_method alias
            query_aliases.append(query_method)
        # there may not be 2 same query aliases
        self.assertEqual(sorted(query_aliases), sorted(set(query_aliases)))

    def test_pm_SearchItemsOfMyCommitteesAndItemsOfMyCommitteesEditable(self):
        '''Test the 'items-of-my-committees' and 'items-of-my-committees-editable'
           adapters that should return a list of items current user is committee
           editor for.'''
        self.changeUser('siteadmin')
        cfg = self.meetingConfig
        cfg.setItemCommitteesStates(['itemcreated'])
        cfg.setItemCommitteesViewStates(['validated'])
        itemTypeName = cfg.getItemTypeName()
        # configure committees editors
        self._setUpCommitteeEditor(cfg)

        # items-of-my-committees
        # first test the generated query, admin is not a committee editor
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='items-of-my-committees')
        self.assertEqual(adapter.query,
                         {'committees_index': {'query': []},
                          'portal_type': {'query': itemTypeName}})
        self.changeUser('pmCreator2')
        self.assertEqual(adapter.query,
                         {'committees_index': {'query': ['committee_1']},
                          'portal_type': {'query': itemTypeName}})
        # now create items and test that the search
        self.changeUser('pmManager')
        editable_item = self.create('MeetingItem', committees=['committee_1'])
        viewable_item = self.create('MeetingItem', committees=['committee_1'])
        # a third item, not viewable by pmCreator2
        self.create('MeetingItem')
        self.validateItem(viewable_item)
        collection = cfg.searches.searches_items.searchitemsofmycommittees
        # pmManager is not a committee editor
        self.failIf(collection.results())
        self.changeUser('pmCreator2')
        res = collection.results()
        self.assertEqual(sorted([brain.UID for brain in res]),
                         sorted([viewable_item.UID(), editable_item.UID()]))

        # items-of-my-committees-editable
        # first test the generated query, admin is not a committee editor
        adapter = getAdapter(cfg,
                             ICompoundCriterionFilter,
                             name='items-of-my-committees-editable')
        self.assertEqual(adapter.query,
                         {'committees_index': {'query': ['committee_1']},
                          'portal_type': {'query': itemTypeName},
                          'review_state': {'query': ('itemcreated', )}})
        collection = cfg.searches.searches_items.searchitemsofmycommitteeseditable
        res = collection.results()
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0].UID, editable_item.UID())

    def test_pm_SearchLivingItems(self):
        '''Test the 'living-items' CompoundCriterion adapter.
           Returns every items that are not decided.'''
        cfg = self.meetingConfig
        collection = cfg.searches.searches_items.searchlivingitems
        self.changeUser('pmManager')
        item = self.create('MeetingItem', decision=self.decisionText)
        self.assertTrue(item.UID() in [brain.UID for brain in collection.results()])
        meeting = self.create('Meeting')
        self.presentItem(item)
        self.freezeMeeting(meeting)
        self.assertTrue(item.UID() in [brain.UID for brain in collection.results()])
        self.decideMeeting(meeting)
        self.assertTrue(item.UID() in [brain.UID for brain in collection.results()])
        self.closeMeeting(meeting)
        self.assertEqual(item.query_state(), 'accepted')
        self.assertTrue(item.query_state() in cfg.getItemDecidedStates())
        self.assertFalse(item.UID() in [brain.UID for brain in collection.results()])

    def test_pm_json_collections_count(self):
        """Test the @@json_collections_count, essentially because it is cached thru
           PMRenderTermView.number_of_items.
           Test also that caching works when using a "myitems" like collection as
           user_id is taken into account in the invalidation key in this case."""
        # disable every showNumberOfItems so no search is enabled by custom profile
        cfg = self.meetingConfig
        for collection in cfg.searches.searches_items.objectValues():
            collection.showNumberOfItems = False
        self.changeUser("pmCreator1")
        view = self.getMeetingFolder().restrictedTraverse("@@json_collections_count")
        self.assertEqual(view(), '{"criterionId": "c1", "countByCollection": []}')
        item = self.create("MeetingItem")
        self.assertEqual(view(), '{"criterionId": "c1", "countByCollection": []}')
        searchmyitems = cfg.searches.searches_items.searchmyitems
        searchmyitems_uid = searchmyitems.UID()
        searchmyitems.showNumberOfItems = True
        self.assertEqual(
            view(),
            '{"criterionId": "c1", "countByCollection": [{"count": 1, "uid": "%s"}]}' % searchmyitems_uid)
        self.changeUser("pmCreator1b", clean_memoize=False)
        view = self.getMeetingFolder().restrictedTraverse("@@json_collections_count")
        self.assertEqual(
            view(),
            '{"criterionId": "c1", "countByCollection": [{"count": 0, "uid": "%s"}]}' % searchmyitems_uid)
        self.changeUser("pmCreator1", clean_memoize=False)
        self.deleteAsManager(item.UID())
        view = self.getMeetingFolder().restrictedTraverse("@@json_collections_count")
        self.assertEqual(
            view(),
            '{"criterionId": "c1", "countByCollection": [{"count": 0, "uid": "%s"}]}' % searchmyitems_uid)


def test_suite():
    from unittest import makeSuite
    from unittest import TestSuite
    suite = TestSuite()
    suite.addTest(makeSuite(testSearches, prefix='test_pm_'))
    return suite
